#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import numpy as np
import geopandas as gpd
import pandas as pd
from solaris.eval.iou import calculate_iou
from pathlib import Path
import os, sys

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def get_iou_df(test_df, truth_df,iou_thresh=0.5, remove_matching_element=True):
    '''
    returns dataframe of test polygons with a column 'match_id' added
    its value for each test polygon is the id of the best match in the truth dataset,
    or -1 if no match was found
    '''
    test_result_list = []

    # we need to use a copy of the truth data, as 
    # matches will be removed during processing.
    copy_truth_df = truth_df.copy(deep=True)

    # insert columns 'match_id' and 'best_iou_score' into the test dataframe
    test_df = test_df.assign(match_id=pd.array(np.full(len(test_df), -1), dtype="int32"))
    test_df = test_df.assign(best_iou_score=pd.array(np.full(len(test_df), np.nan), dtype="float32"))

    test_df = test_df.reset_index()  # make sure indexes pair with number of rows
    for index, row in test_df.iterrows():
        test_poly = row['geometry']

        # get an iou dataframe containing all the rows in the truth data 
        # with polygons that overlap the test polygon
        # this may be empty but iou_df should never be None
        iou_df = calculate_iou(test_poly, copy_truth_df)
        if len(iou_df) == 0:
            row['match_id'] = -1
            row['best_iou_score'] = np.nan
            test_result_list.append(row)
            continue
        max_iou_row = iou_df.loc[iou_df['iou_score'].idxmax(axis=0, skipna=True)]
        best_iou_score = max_iou_row['iou_score']
        if max_iou_row is None or best_iou_score < iou_thresh:
            row['match_id'] = -1
            row['best_iou_score'] = np.nan
            test_result_list.append(row)
            continue
        row['match_id'] = int(max_iou_row.name)
        row['best_iou_score'] = best_iou_score
        if remove_matching_element:
            copy_truth_df.drop(max_iou_row.name, axis=0, inplace=True)
        test_result_list.append(row)

    return  gpd.GeoDataFrame(test_result_list)

def get_iou_accuracy_metrics(test_data, truth_data, size_threshold=None, iou_threshold=0.5):
    '''
    returns dictionary containing accuracy metrics

    test_data: geopandas data frame containing proposal data
    truth_data: geopandas data frame containing truth data
    -- > Both the truth data and test data frames are assumed to have an 'id' field
    
    threshold (float, 0<threshold<1): above which value of IoU a detection is considered to have occurred
    '''

    f_truth_data = truth_data
    f_test_data = test_data

    # apply filtering
    if size_threshold:
        if "area_m2" in truth_data.keys() and "area_m2" in test_data.keys():
            f_truth_data = truth_data[truth_data["area_m2"] >= size_threshold]
            f_test_data = test_data[test_data["area_m2"] >= size_threshold]
        else:
            log.warning("No area_m2 attribute on buildings: cannot apply building size filtering")
    n_truth = len(f_truth_data)
    n_test = len(f_test_data)
    if n_test == 0 or n_truth==0:
        metrics = {
            'n_detections': n_test, 
            'n_truth':n_truth, 
            'true_pos':0,
            'false_pos':n_test,
            'false_neg':n_truth,
            'recall': 0 if n_truth != 0 else None,
            'precision': 0 if n_test != 0 else None,
            'F1': None
        }
        return metrics

    iou_df = get_iou_df(f_test_data, f_truth_data, iou_thresh=iou_threshold)
    assert len(f_truth_data) == n_truth
    true_positives = iou_df[iou_df['match_id'] != -1]
    n_matched = len(true_positives)
    recall = n_matched/n_truth
    precision = n_matched/n_test
    F1 = 2*recall*precision/(recall+precision) if recall+precision != 0 else 0
    metrics = {
        'n_detections': n_test, 
        'n_truth':n_truth, 
        'true_pos':n_matched,
        'false_pos':n_test - n_matched,
        'false_neg':n_truth - n_matched,
        'recall':recall,
        'precision':precision,
        'F1':F1
    }
    return metrics



#################################################################
#
# created for ramp project, August 2022
#
#################################################################



import numpy as np

def get_best_model_value_and_epoch(model_history):
    '''
    argument: the history object returned by model.fit.
    '''
    val_acc = model_history.history["val_sparse_categorical_accuracy"]
    best_epoch = np.argmax(val_acc)
    best_value = val_acc[best_epoch]
    return best_epoch, best_value



#################################################################
#
# created for ramp project, August 2022
#
#################################################################


import glob, os, re
from pathlib import Path
import csv

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def check_create_dir(dirname):
    if not os.path.exists(dirname):
        log.info('Creating directory: {dirname}')
        os.makedirs(dirname, exist_ok=True)

# Used to check whether directory paths are given in the arguments
def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise NotADirectoryError(string)
        
def get_num_files(dirname, pattern = "*.tif"):
    ''' 
    get the number of TIFF files in a directory.
    '''
    return len(glob.glob(os.path.join(dirname, pattern)))

# this is here because a lot of older notebooks import this function but don't use it
def get_experiment_db():
    pass


def log_experiment_to_file(
    the_experiment, 
    the_filename=os.path.join(os.environ['RAMP_HOME'],'ramp-code/ramp_experiments.csv')
    ):
    '''
    log an experiment to an existing or new csv file.
    param: dict the_experiment: key-value store of information to be logged about the experiment. 
    param: str the_filename: file to log the experiment to.  
    '''

    csvpath = Path(the_filename)    
    if csvpath.is_file():
        
        # file exists, open and read
        newrows = []
        fieldnames = []
        with open(str(csvpath),'r') as csvfile:
            reader = csv.DictReader(csvfile)

            ## bugfix: add new fieldnames to the experiment csv
            fieldnames = reader.fieldnames
            for row in reader:
                newrows.append(row)

        expkeys = the_experiment.keys()
        for key in expkeys: 
            if key not in fieldnames:
                fieldnames.append(key)

        newrows.append(the_experiment)

        with open(str(csvpath), 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(newrows)

    else:
        # file does not exist
        with open(str(csvpath), 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=the_experiment.keys())
            writer.writeheader()
            writer.writerows([the_experiment])

    return

def get_chip_number_sn2(the_filename):
    '''
    extract chip number from any data filename in Spacenet 2 format
    '''
    import re
    import os.path

    chipnum_pattern = re.compile(r"(?P<numbers>[0-9]+)$")
    extensions = set('.tiff .tif .json .geojson'.split())
    fbase, ext = os.path.splitext(the_filename)

    if ext not in extensions:
        return None 

    chip_match = chipnum_pattern.search(fbase)
    if chip_match is None:
        log.warning(f"Unhandled filename: {the_filename}")
        return None

    chipnum = int(chip_match.group("numbers"))
    return chipnum


from . import augmentation_constructors
from . import callback_constructors
from . import loss_constructors
from . import metric_constructors
from . import model_constructors
from . import optimizer_constructors


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################



from .img_utils import create_gdal_dataset_from_mask_tensor
from osgeo import gdal, ogr, osr
import numpy as np

def onehot_to_sparse_multimask(onehot_mask):
    """
    Change a one-hot (channels-last) mask to a sparse-encoded mask
    onehot_mask: numpy ndarray of shape [H,W, C] with values 0,1
    sparse_mask: numpy ndarray of shape [H,W,1] with values 0,...,C-1 
    """
    sparse_mask = np.zeros(onehot_mask.shape[:2], dtype=np.uint8)
    sparse_mask[np.where(onehot_mask[:,:,0] > 0)] = 1
    sparse_mask[np.where(onehot_mask[:,:,1] > 0)] = 2
    sparse_mask[np.where(onehot_mask[:,:,2] > 0)] = 3
    return np.expand_dims(sparse_mask, axis=-1)

def sparse_multimask_to_onehot_mask(num_classes, sparse_mask):
    """
    NOT TESTED and probably buggy
    Change a sparse-encoded to a onehot (channels-last) mask
    num_classes: C
    sparse_mask: numpy ndarray of shape [H,W,1] with values in 0,...,C-1 
    onehot_mask: numpy ndarray of shape [H,W, C] with values 0,1
    """

    raise NotImplementedError

    # get H and W
    dims = sparse_mask.shape[:2]

    # get rid of 3rd trivial mask dimension if present
    if len(sparse_mask.shape) == 3:
        sparse_mask = np.squeeze(sparse_mask, axis=-1)

    indices = np.indices(sparse_mask.shape)

    # construct onehot mask
    onehot_mask = np.zeros((*dims, num_classes), dtype = np.uint8)
    onehot_mask[indices[0],indices[1], sparse_mask]=1
    return onehot_mask

def binary_mask_from_multichannel_mask(multimask):
    '''
    create a binary building mask from a multichannel prediction.
    :param ndarray multimask: 4-channel, sparse-encoded multichannel mask
    (classes: background:0, building:1, boundary:2, close-contact-point:3)
    '''

    # aggregate building points, remaining points are 0
    # 20220808: can't include boundary points, in tight places you get blobs.
    foregd_mask = (multimask == 1)
    return foregd_mask.astype(np.uint8)


def binary_mask_from_sdt_mask(sdt_mask, bldg_lower_thresh):
    '''
    create a binary building mask from a prediction based on the Truncated Signed Distance Transform.
    Right now this process is a simple thresholding. 
    We'll add more post-processing functionality here as necessary. 

    :param ndarray sdt_mask: int32 truncated signed distance transform mask.
    :param int bldg_lower_thresh: threshold below which the mask pixel will be set to 0. 
    '''
    mask = sdt_mask >= bldg_lower_thresh
    return mask.astype('uint8')

def binary_mask_to_geojson(mask, reference_ds, geojson_path):
    '''

    Polygonize a binary mask to geojson file.
    I wanted to just print it out to a string, but can't figure out how to do this:
    OGR triumphs again.

    :param ndarray mask: numpy ndarray containing the binary byte mask to polygonize.
    :param reference_ds: gdal dataset containing the georeferencing to use for the output geojson. 
    :param geojson_path: path to geojson file for exported data.
    '''

    # insert binary mask into an in-memory gdal dataset with the same georeferencing as the reference_ds.
    # this should be an img_utils function call. 
    mask_ds = create_gdal_dataset_from_mask_tensor(reference_ds, mask)
    
    # create a geojson layer to polygonize into.
    # this as usual is a pita in OGR. 
    drv = ogr.GetDriverByName('GeoJSON')
    vec_ds = drv.CreateDataSource(geojson_path)
    dst_layername = 'mask'
    prj = reference_ds.GetProjection()
    
    srs = osr.SpatialReference(wkt=prj)
    layer = vec_ds.CreateLayer(dst_layername, srs = srs, geom_type = ogr.wkbPolygon )
    
    # gdal.Polygonize will want to write the pixel value of the connected component
    # into an attribute in the layer. You need to create this attribute.
    # if you don't, the code will run anyway, but OGR will complain. 
    # this is very poorly documented online. 
    # explanation here: https://gdal.org/api/gdal_alg.html#_CPPv414GDALPolygonize15GDALRasterBandH15GDALRasterBandH9OGRLayerHiPPc16GDALProgressFuncPv
    pixValueField = ogr.FieldDefn( 'pixel_value', ogr.OFTInteger )
    layer.CreateField( pixValueField )

    # get the mask, and call polygonize. 
    # no one online seems clear on what that 4th parameter does. 
    band = mask_ds.GetRasterBand(1)

    # that second parameter causes gdal to use the same band for a mask, to determine which connected 
    # regions to polygonize. 0 areas (background!) are not polygonized. Otherwise, if mask is None, 
    # the background gets polygonized, not just the buildings.
    # gdal.Polygonize(band, None, layer, 0, [])
    gdal.Polygonize(band, band, layer, 0, [])
    # gdal.Polygonize(band, None, layer, 0, [])

    # and brilliant: here's how you save and close the geojson file you created. 
    vec_ds = None
    return
    



def sdt_mask_to_geojson(sdt_mask, reference_ds, geojson_path, bldg_lower_thresh=3):
    '''
    convert truncated sdt mask to a geojson and writes it to a file.

    :param ndarray sdt_mask: byte mask derived from truncated signed distance transform.
    :param GdalDataset reference_ds: dataset from which to extract georeferencing (usually the geotiff image chip).
    :param int bldg_lower_threshold: threshold below which a mask pixel is considered background and set to 0.
    '''

    # create binary mask. 
    bin_mask = binary_mask_from_sdt_mask(sdt_mask, bldg_lower_thresh)
    binary_mask_to_geojson(bin_mask, reference_ds, geojson_path)
    return



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import segmentation_models as sm


BACKBONE = 'efficientnetb0'
CLASSES = ['building']

# This call appears to retrieve the names of some of 
# the layers of the backbone model. 
# preprocess_input = sm.get_preprocessing(BACKBONE)

def get_effunet(backbone=BACKBONE,
                classes=CLASSES, 
                encoder_freeze=False):

    # define network parameters

    n_classes = len(classes)
    # n_classes = 1 if len(classes) == 1 else (len(classes) + 1)  # case for binary and multiclass segmentation
    activation = 'sigmoid' if n_classes == 1 else 'softmax'

    #create model
    model = sm.Unet(backbone, classes=n_classes, activation=activation, encoder_freeze=encoder_freeze)
    return model





#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################



# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

import numpy as np
import cv2
import tensorflow as tf

def get_edge_image(mask_tensor):
    '''
    takes a mask tensor with shape (H,W) or (H,W,1)
    returns a (H,W)-shaped float32 numpy array with edge points marked as 1.
    '''
    # mask tensor will frequently have shape (H,W,1)
    # we need (H,W)
    mask = np.squeeze(mask_tensor.numpy())

    # even though the edge_image is 0,1-valued,
    # it has to be a float32 or gaussian blurring will break. 
    edge_image = np.zeros(mask.shape, dtype=np.float32 )

    # contours is a dataset of boundary points
    # CHAIN_APPROX_NONE means all points are retained 
    # for burning into the edge image 
    contours, t2 = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

    for contour in contours:
        
        # contour has dimensions: [#pts in contour, 1, 2]
        # get rid of the extra 1...
        pts_array = np.squeeze(contour, 1)

        # turn pts_array of shape [#pts, 2] into a list of x,y pairs
        pts = [pts_array[ii,:] for ii in range(pts_array.shape[0])]

        # mark every border point in the edge image.
        for pt in pts:
            edge_image[pt[1], pt[0]] = 1
    

    return edge_image

def get_weight_image(mask_tensor, sigma, myconst):
    '''
    takes uint8 binary mask tensor of dimensions (H,W) or (H,W,1)
    sigma is scale length for gaussian kernel in pixels
    myconst is a multiplicative constant applied to the weight image
    returns float64 weight image of dimensions (H,W)
    '''
    edge_image = get_edge_image(mask_tensor)
    
    # important undocumented note: the blur image will have the same datatype as the edge image. 
    # so if the edge image is uint8, the blur image will be 0. 
    blur_image = cv2.GaussianBlur(edge_image, 
                                    ksize=(0,0), 
                                    sigmaX=sigma, 
                                    sigmaY=sigma, 
                                    borderType=cv2.BORDER_CONSTANT)

    return myconst*blur_image

def batch_get_weight_image(mask_tensor_batch, sigma, myconst):
    '''
    runs get_weight_image over a batch of mask tensors.
    batch dimension is first.
    returns numpy array of shape (B, H, W).
    '''
    
    # this will be (B,H,W,1) for sparse-encoded masks
    (B, H, W, C) = mask_tensor_batch.shape
    batch_weight_img = np.zeros([B, H, W])

    for ii in range(B):
        mask_tensor_ii = mask_tensor_batch[ii,...]
        batch_weight_img[ii,...] = get_weight_image(mask_tensor_ii, sigma, myconst)

    return batch_weight_img



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

from tensorflow import keras
from tensorflow.keras import layers


def get_model_1_chollet_unet(img_size, num_classes):
    '''
    This was once the baseline model for the ramp project. It's simple and susceptible to overfit. 
    This model is a unet-like model, written by Chollet in his segmentation example at this webpage:
    https://keras.io/examples/vision/oxford_pets_image_segmentation/.
    It's not a transfer model. All parameters are trainable.
    '''
    inputs = keras.Input(shape=img_size + (3,))

    ### [First half of the network: downsampling inputs] ###

    # Entry block
    x = layers.Conv2D(32, 3, strides=2, padding="same")(inputs)
    x = layers.BatchNormalization()(x)
    x = layers.Activation("relu")(x)

    previous_block_activation = x  # Set aside residual

    # Blocks 1, 2, 3 are identical apart from the feature depth.
    for filters in [64, 128, 256]:
        x = layers.Activation("relu")(x)
        x = layers.SeparableConv2D(filters, 3, padding="same")(x)
        x = layers.BatchNormalization()(x)

        x = layers.Activation("relu")(x)
        x = layers.SeparableConv2D(filters, 3, padding="same")(x)
        x = layers.BatchNormalization()(x)

        x = layers.MaxPooling2D(3, strides=2, padding="same")(x)

        # Project residual
        residual = layers.Conv2D(filters, 1, strides=2, padding="same")(
            previous_block_activation
        )
        x = layers.add([x, residual])  # Add back residual
        previous_block_activation = x  # Set aside next residual

    ### [Second half of the network: upsampling inputs] ###

    for filters in [256, 128, 64, 32]:
        x = layers.Activation("relu")(x)
        x = layers.Conv2DTranspose(filters, 3, padding="same")(x)
        x = layers.BatchNormalization()(x)

        x = layers.Activation("relu")(x)
        x = layers.Conv2DTranspose(filters, 3, padding="same")(x)
        x = layers.BatchNormalization()(x)

        x = layers.UpSampling2D(2)(x)

        # Project residual
        residual = layers.UpSampling2D(2)(previous_block_activation)
        residual = layers.Conv2D(filters, 1, padding="same")(residual)
        x = layers.add([x, residual])  # Add back residual
        previous_block_activation = x  # Set aside next residual

    # Add a per-pixel classification layer
    outputs = layers.Conv2D(num_classes, 3, activation="softmax", padding="same")(x)

    # Define the model
    model = keras.Model(inputs, outputs)
    return model



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing 
import os
from osgeo import gdal, gdalnumeric as gdn
import numpy as np
import matplotlib.pyplot as plt
from tensorflow.python.ops.gen_batch_ops import batch
from ..utils.ramp_exceptions import GdalReadError
import random

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

###########################################################################################

def tf_gdal_get_image_tensor(image_path):    
  '''
  Reads a satellite image geotiff using gdal, and converts to a numpy tensor.
  Image geotiffs have 3 channels (RGB) and are of type float32.

  
  Channels go last in the index order for the output tensor, as is standard for tensorflow.

  :param tf.StringTensor image_path: path to the image file, as a tensorflow String Tensor.
  '''

  its_dtype = "float32"
  ds  = gdal.Open(image_path.numpy())
  if not ds:
      log.error(f"GDAL could not read {image_path.numpy()}")
      raise GdalReadError(f"Could not read {image_path.numpy()}")
  bands = [ds.GetRasterBand(i) for i in range(1, ds.RasterCount + 1)]
  cols = ds.RasterXSize 
  rows = ds.RasterYSize 
    
  # create a channels_last array -- rows, columns
  the_array = np.zeros((rows, cols, ds.RasterCount), its_dtype)
  
  for bnum, the_band in enumerate(bands):
    from_img = gdn.BandReadAsArray(the_band)
    to_img = the_array[:,:,bnum]
    np.copyto(to_img, from_img)
    
  # normalize the float image to [0,1]
  return the_array/np.max(the_array)


def tf_gdal_get_image_tensor_with_name(image_path):

  return tf_gdal_get_image_tensor(image_path), image_path

###########################################################################################

def tf_gdal_get_mask_tensor(mask_path):
  '''
  Reads a mask geotiff image using gdal, and converts to a numpy tensor.
  Masks have datatype uint8, and a single channel.
  mask_path: path to a gdal-readable image.
  
  Channels go last in the index order for the output tensor, as is standard for tensorflow.

  :param tf.StringTensor mask_path: path to the image file, as a tensorflow String Tensor.
  '''

  its_dtype = "uint8"

  ds  = gdal.Open(mask_path.numpy())
  if not ds:
    log.error(f"GDAL could not read {mask_path.numpy()}")
    raise GdalReadError(f"Could not read {mask_path.numpy()}")
  bands = [ds.GetRasterBand(i) for i in range(1, ds.RasterCount + 1)]
  cols = ds.RasterXSize 
  rows = ds.RasterYSize 
    
  # create a channels_last array -- rows, columns
  the_array = np.zeros((rows, cols, ds.RasterCount), its_dtype)
  
  for bnum, the_band in enumerate(bands):
    from_img = gdn.BandReadAsArray(the_band)
    to_img = the_array[:,:,bnum]
    np.copyto(to_img, from_img)
  return the_array

def tf_gdal_get_mask_tensor_with_name(mask_path):

  return tf_gdal_get_mask_tensor(mask_path), mask_path

###########################################################################################

def resize_it(an_image, original_shape, new_size, the_method):
    '''
    Used with 'map' to resize images in the data pipeline.
    an_image: an image array of shape 'original shape'.
    original_shape: the shape of the input image array.
    new_size: the new dimensions (width, height) of the output image array. 
        The number of channels is unchanged, so is not specified in 'new_size'.
    the_method: method specified for interpolation. 'nearest' for masks, 'bilinear' for images. 
    '''

    # this call to set_shape works around a bug in tensorflow
    # the bug is described in this article: 
    # https://stackoverflow.com/questions/62957726/i-got-value-error-that-image-has-no-shape-while-converting-image-to-tensor-for-p
    an_image.set_shape(original_shape)
    return tf.image.resize(an_image, new_size, method=the_method)

def resize_it_with_name(an_image, original_shape, new_size, the_method, pathname):

  return resize_it(an_image, original_shape, new_size, the_method), pathname


  # AUGMENTATION ADDITIONS
  ###########################################################################################

# AUG
import albumentations as A
from functools import partial



### APPLY AUGMENTATION WORKAROUND

# This is a function that prepackages another function that actually applies augmentation.
# This workaround is needed because I can't pass img_shape and transforms to tf.numpy_function.

def get_apply_augmentation_function(img_shape, transforms):
  '''
  Function that prepackages a function that applies augmentation.
  Needed because I can't pass img_shape and transforms to tf.numpy_function.
  '''

  # define prepackaged function to apply augmentation, using the given albumentation transform.
  def apply_augmentation(image, mask):
    '''
    Apply augmentation using the albumentations library.
    image: ndarray (image tensor).
    mask: ndarray (mask tensor)

    Packaged parameters:
    img_shape: shape to use in call to tf.image.resize, following the transform.
    #todo: can you replace this with a call to resize_it?
    transforms: the augmentation object returned from call to albumentations.Compose().
    returns: augmented image and mask.   

    Code reference page: https://albumentations.ai/docs/examples/tensorflow-example/
    Note: certain augmentations are applied only to the mask.
    See the bottom of this webpage: https://github.com/albumentations-team/albumentations_examples/blob/master/notebooks/example_kaggle_salt.ipynb.  
    '''
    data = {"image":image, "mask":mask}
    
    if transforms is not None:
      aug_data = transforms(**data)
      aug_img = aug_data["image"]
      aug_mask = aug_data["mask"]
    else:
      aug_img = image
      aug_mask = mask

    ## DEBUG
    # check if mask was transformed
    # changed = aug_mask != mask
    # print(f'Mask was transformed: {changed.any()}')

    # not needed -- aug_img is already normalized to the range 0-1
    # renormalization might be needed once contrast/brightness shifts are added to augmentations
    # aug_img = tf.cast(aug_img/255.0, tf.float32)
    aug_mask = tf.cast(aug_mask, tf.uint8)
    aug_img = tf.image.resize(aug_img, size=img_shape)
    aug_mask = tf.image.resize(aug_mask, size=img_shape, method='nearest')
    return aug_img, aug_mask

  # return the prepackaged function
  return apply_augmentation
### END APPLY AUGMENTATION WORKAROUND

def process_data(image, mask, img_shape, transforms):
  '''
  Code reference page: https://albumentations.ai/docs/examples/tensorflow-example/
  '''
  apply_aug_fn = get_apply_augmentation_function(img_shape, transforms)
  aug_img, aug_mask = tf.numpy_function(func = apply_aug_fn, inp = [image, mask], Tout=[tf.float32, tf.uint8])
  #print(aug_img.dtype, aug_mask.dtype)
  return aug_img, aug_mask

def reset_shapes(image, mask, image_shape):
  image.set_shape([*image_shape, 3])
  mask.set_shape([*image_shape,1])
  return image, mask

  # END AUGMENTATION ADDITIONS
  ###########################################################################################


def training_batches_from_gtiff_dirs(
    image_file_dir,
    mask_file_dir,
    batch_size, 
    input_image_size,
    output_image_size,
    transforms = None):
    '''
    image_file_dir: a directory containing geotiff image files to be used in training.
    mask_file_dir: a directory containing matching geotiff mask files to be used in training.
    batch_size: batch size, a positive integer.
    input_image_size: dimensions of the input images in the form (rows, columns).
    output_image_size: dimensions of the output images in the form (new_rows, new_columns).
    tfgseed: Tensorflow global random seed for reproducible results. 
    transforms: Albumentations object for augmenting training data. None == no augmentation (default).

    Matching images and masks are assumed to have names of the form:
    unique_string.tif, same_unique_string.mask.tif. 
    
    Examples of matching image and mask: 
    "training_img909.tif",
    "training_img909.mask.tif".

    Example input image size: (650, 650).
    Example output image size: (512, 512).

    Returns: a tf.data.Dataset containing batches of matching images and masks, \
      suitable for using in model.fit. 
   '''

    tf_global_seed = 20220607
    py_seed = 303041

    # set this for reproducibility in file listings
    tf.random.set_seed(tf_global_seed)

    # set this for reproducibility in augmentations
    random.seed(py_seed)

    image_pat = os.path.join(image_file_dir, '*.tif')
    mask_pat = os.path.join(mask_file_dir, '*.mask.tif')


    # giving these the same op-level random seed causes them to be listed in the same order.
    image_ds = tf.data.Dataset.list_files(image_pat, seed=333)
    mask_ds = tf.data.Dataset.list_files(mask_pat, seed=333)

    # get the number of training images and masks. Since they match, the numbers should be the same. 
    n_images = tf.data.experimental.cardinality(image_ds)
    n_masks = tf.data.experimental.cardinality(mask_ds)
    assert n_images==n_masks, f"Number of images: {n_images} | Number of masks: {n_masks}"
    
    # Define functions to use for loading gdal images in the data pipeline.
    # These are just Tensorflow wrappers around 
    # the gdal_get_image_tensor and gdal_get_mask_tensor functions.
    tf_load_image_fn = lambda imgname: tf.py_function(func = tf_gdal_get_image_tensor, 
                              inp=[imgname], 
                              Tout=[tf.float32])

    tf_load_mask_fn = lambda maskname: tf.py_function(func = tf_gdal_get_mask_tensor,
                                                   inp = [maskname],
                                                    Tout = [tf.uint8])
    
    # resize the images and masks to the desired size.
    images = image_ds.map(tf_load_image_fn).map(lambda x: resize_it(x, [*input_image_size, 3], output_image_size, 'bilinear'))
    masks = mask_ds.map(tf_load_mask_fn).map(lambda x: resize_it(x, [*input_image_size, 1], output_image_size,'nearest'))

    # zip them together so the resulting dataset puts out a data element of the form (image, matching mask).
    zip_pairs = tf.data.Dataset.zip((images, masks), name=None)

    len_train = n_images
    buffer_size = len_train

    data_batches = None
    
    if transforms is not None:
      data_batches = (
      zip_pairs
      .cache()
      .shuffle(buffer_size)
      # START augmentation 
      # the next two lines apply the augmentation. 
      # The dataset loses its shape after applying a tf.numpy_function, 
      # so it's necessary to call reset_shapes for the sequential model and when inheriting from the model class.
      # See 'restoring dataset shapes' in https://albumentations.ai/docs/examples/tensorflow-example/.
      .map(partial(process_data, img_shape=output_image_size, transforms=transforms),
                      num_parallel_calls=tf.data.AUTOTUNE)
      .map(partial(reset_shapes, image_shape=output_image_size), 
                      num_parallel_calls=tf.data.AUTOTUNE)
      # END augmentation
      .batch(batch_size)
      .repeat()
      .prefetch(buffer_size=tf.data.AUTOTUNE)
      )


    else:
      data_batches = (
        zip_pairs
        .cache()
        .shuffle(buffer_size)
        .batch(batch_size)
        .repeat()
        .prefetch(buffer_size=tf.data.AUTOTUNE)
        )

    return data_batches
    
###########################################################################################


def test_batches_from_gtiff_dirs(
    image_file_dir,
    mask_file_dir,
    batch_size,
    input_image_size,
    output_image_size
    ):
      '''

      Data generator for test batches.
      This data generator does not shuffle, cache, repeat, or augment -- it just delivers all
      test image batches in the same order every time.

      See the notes for 'train_batch_from_gtiff_dirs' for more details.
      '''

      image_pat = os.path.join(image_file_dir, '*.tif')
      mask_pat = os.path.join(mask_file_dir, '*.tif')


      # giving these the same random seed causes them to be listed in the same order.
      image_ds = tf.data.Dataset.list_files(image_pat, seed=333)
      mask_ds = tf.data.Dataset.list_files(mask_pat, seed=333)

      # get the number of training images and masks. Since they match, the numbers should be the same.
      # todo: make the data generator end appropriately if the numbers don't match.
      n_images = tf.data.experimental.cardinality(image_ds)
      n_masks = tf.data.experimental.cardinality(mask_ds)
      assert n_images==n_masks, f"Num images: {n_images}, Num masks: {n_masks} "

      # Define functions to use for loading gdal images in the data pipeline.
      # These are just Tensorflow wrappers around
      # the gdal_get_image_tensor and gdal_get_mask_tensor functions.
      tf_load_image_fn = lambda imgname: tf.py_function(func = tf_gdal_get_image_tensor,
                                inp=[imgname],
                                Tout=[tf.float32])

      tf_load_mask_fn = lambda maskname: tf.py_function(func = tf_gdal_get_mask_tensor,
                                                    inp = [maskname],
                                                      Tout = [tf.uint8])

      # resize the images and masks to the desired size.
      images = image_ds.map(tf_load_image_fn).map(lambda x: resize_it(x, [*input_image_size, 3], output_image_size, 'bilinear'))
      masks = mask_ds.map(tf_load_mask_fn).map(lambda x: resize_it(x, (*input_image_size, 1), output_image_size,'nearest'))

      # zip them together so the resulting dataset puts out a data element of the form (image, matching mask).
      zip_pairs = tf.data.Dataset.zip((images, masks), name=None)

      len_train = n_images
      buffer_size = len_train

      data_batches = (
        zip_pairs.batch(batch_size)
      )

      return data_batches

###################################################################

def test_batches_from_gtiff_dirs_with_names(
    image_file_dir,
    mask_file_dir,
    batch_size,
    input_image_size,
    output_image_size
    ):
      '''

      Data generator for test batches.
      This data generator does not shuffle, cache, repeat, or augment -- it just delivers all
      test image batches in the same order every time.

      See the notes for 'train_batch_from_gtiff_dirs' for more details.
      '''

      image_pat = os.path.join(image_file_dir, '*.tif')
      mask_pat = os.path.join(mask_file_dir, '*.tif')


      # giving these the same random seed causes them to be listed in the same order.
      image_ds = tf.data.Dataset.list_files(image_pat, seed=333)
      mask_ds = tf.data.Dataset.list_files(mask_pat, seed=333)

      # get the number of training images and masks. Since they match, the numbers should be the same.
      # todo: make the data generator end appropriately if the numbers don't match.
      n_images = tf.data.experimental.cardinality(image_ds)
      n_masks = tf.data.experimental.cardinality(mask_ds)
      assert n_images==n_masks, f"Num images: {n_images}, Num masks: {n_masks} "

      # Define functions to use for loading gdal images in the data pipeline.
      # These are just Tensorflow wrappers around
      # the gdal_get_image_tensor and gdal_get_mask_tensor functions.
      tf_load_image_fn = lambda imgname: tf.py_function(func = tf_gdal_get_image_tensor_with_name,
                                inp=[imgname],
                                Tout=[tf.float32, tf.string])

      tf_load_mask_fn = lambda maskname: tf.py_function(func = tf_gdal_get_mask_tensor_with_name,
                                                    inp = [maskname],
                                                      Tout = [tf.uint8, tf.string])

      # resize the images and masks to the desired size.
      images = image_ds.map(tf_load_image_fn).map(lambda x, pathname: resize_it_with_name(x, [*input_image_size, 3], output_image_size, 'bilinear', pathname))
      masks = mask_ds.map(tf_load_mask_fn).map(lambda x, pathname: resize_it_with_name(x, (*input_image_size, 1), output_image_size,'nearest', pathname))

      # zip them together so the resulting dataset puts out a data element of the form (image, matching mask).
      zip_pairs = tf.data.Dataset.zip((images, masks), name=None)

      len_train = n_images
      buffer_size = len_train

      data_batches = (
        zip_pairs.batch(batch_size)
      )

      return data_batches
      

###########################################################################################
## AUGUST 3rd 2022: modifications for using weighted loss functions
## for details on this approach to using class weights in loss computation see this site:
## https://www.tensorflow.org/tutorials/images/segmentation#optional_imbalanced_classes_and_class_weights 
###########################################################################################

def add_class_weights(chip, label, class_weights):
  ''' 
  create a class weight image with the same shape as the 'label' tensor,
  and return it along with the chip and label. 
  class_weights: a float tensorflow (not numpy) array of length equal to the number of classes,
  with weight values to be applied in the loss computation for each class.
  '''

  # normalize the class weights so they sum to 1
  class_weights = class_weights/tf.reduce_sum(class_weights)

  # create the weight tensor. 'gather' works by creating a tensor the same size as the 
  # 'label' object, and using the class index at each pixel of the label to index into the class_weights values.
  weight_tensor = tf.gather(class_weights, tf.cast(label, tf.int32))
  return chip, label, weight_tensor


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################



from tensorflow.keras.callbacks import LambdaCallback
from tensorflow.keras import backend as K
import matplotlib.pyplot as plt
import numpy as np
import tempfile

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

import os, sys
from pathlib import Path

RAMP_HOME = os.environ["RAMP_HOME"]

class LearningRateFinder:
    def __init__(self, model, cfg):
        self.model = model
        lrf_cfg = cfg["lrfinder"]
        self.stopFactor = lrf_cfg["stopFactor"] # usually 4
        self.beta = lrf_cfg["beta"]
        self.startLR = lrf_cfg["startLR"]
        self.endLR = lrf_cfg["endLR"]
        self.lrfind_plot_path = str(Path(RAMP_HOME)/lrf_cfg["lrfind_plot_path"])

        # other cfgs
        self.n_epochs = cfg["num_epochs"]
        self.n_train = cfg["runtime"]["n_training"]
        self.batch_size = cfg["batch_size"]
        self.lrs = []
        self.losses = []
        self.lrMult = 1
        self.avgLoss = 0
        self.bestLoss = 1e9
        self.batchNum = 0
        self.weightsFile = None
    
    def reset(self):
        self.lrs = []
        self.losses = []
        self.lrMult = 1
        self.avgLoss = 0
        self.bestLoss = 1e9
        self.batchNum = 0
        self.weightsFile = None

    
    def on_batch_end(self, batch, logs):
        # grab the current learning rate and add log it to the list of
		# learning rates that we've tried
        lr = K.get_value(self.model.optimizer.lr)
        self.lrs.append(lr)
	
    	# grab the loss at the end of this batch, increment the total
		# number of batches processed, compute the average average
		# loss, smooth it, and update the losses list with the
		# smoothed value
        l = logs["loss"]
        self.batchNum += 1
        self.avgLoss = (self.beta * self.avgLoss) + ((1 - self.beta) * l)
        smooth = self.avgLoss / (1 - (self.beta ** self.batchNum))
        self.losses.append(smooth)

		# compute the maximum loss stopping factor value
        stopLoss = self.stopFactor * self.bestLoss

		# check to see whether the loss has grown too large
        if self.batchNum > 1 and smooth > stopLoss:
            # stop training and return from the method
            self.model.stop_training = True
            return

		# check to see if the best loss should be updated
        if self.batchNum == 1 or smooth < self.bestLoss:
            self.bestLoss = smooth

		# increase the learning rate
        lr *= self.lrMult
        K.set_value(self.model.optimizer.lr, lr)


    def find(self, train_batches):
        # reset our class-specific variables
        self.reset()

        epochs = self.n_epochs
        batch_size = self.batch_size
        startLR = self.startLR
        endLR = self.endLR

        # computed parameters
        steps_per_epoch = self.n_train // self.batch_size
        num_batch_updates = epochs * steps_per_epoch
        self.lrMult = (endLR/startLR) ** (1.0/num_batch_updates)
        log.info(f"startLR/endLR: {startLR}/{endLR}")
        log.info(f"num_batch_updates: {num_batch_updates}")
        log.info(f"lrMult: {self.lrMult}")

        # save the initial model weights
        self.weightsFile = tempfile.mkstemp()[1]
        self.model.save_weights(self.weightsFile)

        # grab the *original* learning rate (so we can reset it
		# later), and then set the *starting* learning rate
        origLR = K.get_value(self.model.optimizer.lr)
        K.set_value(self.model.optimizer.lr, startLR)

        # construct the callback that will be called at the end of each
		# batch, enabling us to increase our learning rate as training
		# progresses
        callbacks_list = []
        callbacks_list.append(LambdaCallback(on_batch_end=lambda batch, logs:
			self.on_batch_end(batch, logs)))

        #### DO THE TRAINING ####

        history = self.model.fit(
            train_batches, 
            epochs=epochs, 
            steps_per_epoch=steps_per_epoch, 
            callbacks=callbacks_list)

        # restore the original model weights and learning rate
        # question, do we need this if we aren't training in this script?
		# self.model.load_weights(self.weightsFile)
		# K.set_value(self.model.optimizer.lr, origLR)

    def plot_loss(self, title="", skipBegin=10, skipEnd=1):
        # grab the learning rate and losses values to plot
        lrs = self.lrs[skipBegin:-skipEnd]
        losses = self.losses[skipBegin:-skipEnd]
		
        # plot the learning rate vs. loss
        plt.plot(lrs, losses)
        plt.xscale("log")
        plt.xlabel("Learning Rate (Log Scale)")
        plt.ylabel("Loss")
		# if the title is not empty, add it to the plot
        if title != "":
        	plt.title(title)

    


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
import os, sys
import tensorflow as tf
import keras
import keras.backend as K
import datetime
from pathlib import Path
import numpy as np

RAMP_HOME = os.environ["RAMP_HOME"]

import tensorflow.summary as tf_summary



from ..data_mgmt.clr_callback import CyclicLR
from . import metric_constructors

################################## HELPER FUNCTIONS

def get_tb_log_dir(cfg):
    timestamp = cfg["timestamp"]
    log_dir = str(Path(RAMP_HOME)/ cfg["tensorboard"]["tb_logs_dir"] / f"log-{timestamp}")
    Path(log_dir).mkdir(parents=True, exist_ok=True)
    return log_dir

def get_accuracy_name(cfg):
    # get the accuracy name from the first metric listed
    assert cfg["metrics"]["use_metrics"], "metrics are not defined"
    get_metrics_fn_names = cfg["metrics"]["get_metrics_fn_names"]
    get_metrics_fn_name = get_metrics_fn_names[0]
    get_metrics_fn = getattr(metric_constructors, get_metrics_fn_name)
    metrics_fn = get_metrics_fn(cfg)
    return metrics_fn.name

################################## CALLBACKS
def get_tb_callback_fn(cfg):
    '''
    returns tensorboard logging callback function
    '''
    log_dir = get_tb_log_dir(cfg)
    tb_callback_parms = cfg["tensorboard"]["tb_callback_parms"]
    return tf.keras.callbacks.TensorBoard(log_dir, **tb_callback_parms)

def get_model_checkpt_callback_fn(cfg):
    '''
    return model checkpt callback fn
    '''
    accuracy_name = get_accuracy_name(cfg)
    val_acc_name = f"val_{accuracy_name}"

    timestamp = cfg["timestamp"]
    models_dir = str(Path(RAMP_HOME)/ cfg["model_checkpts"]["model_checkpts_dir"]/timestamp)
    Path(models_dir).mkdir(parents=True, exist_ok=True)
    model_checkpt_callback_parms = cfg["model_checkpts"]["model_checkpt_callback_parms"]
    Path(models_dir).mkdir(parents=True, exist_ok=True)
    
    # Note to self: '{{' in an f-string escapes the curly braces.
    model_filename = os.path.join(models_dir, f'model_{timestamp}_'+ f'{{epoch:03d}}_{{{val_acc_name}:.3f}}.tf')

    return keras.callbacks.ModelCheckpoint(model_filename, 
                                                monitor = val_acc_name,
                                                **model_checkpt_callback_parms
                                                )

def get_prediction_logging_fn(model, cfg):

    log_dir = get_tb_log_dir(cfg)
    
    #### define data directories ####
    val_img_dir = Path(RAMP_HOME) / cfg["datasets"]["val_img_dir"]
    val_mask_dir = Path(RAMP_HOME) / cfg["datasets"]["val_mask_dir"]
    batch_size = cfg["batch_size"]
    input_img_shape = cfg["input_img_shape"]
    output_img_shape = cfg["output_img_shape"]

    log.debug(f"logging predicted masks to logging directory: {log_dir}")
    log.debug(f"model class: {type(model)}")
    log.debug(f"test image directory: {val_img_dir}")
    log.debug(f"test mask directory: {val_mask_dir}")
    log.debug(f"batch_size: {batch_size}")
    log.debug(f"input image shape: {input_img_shape}")
    log.debug(f"output image shape: {output_img_shape}")

    def log_predictions(epoch, logs):
        '''
        callback function for logging predictions after every epoch.
        '''

        from ..data_mgmt.data_generator import test_batches_from_gtiff_dirs_with_names
        from ..data_mgmt.display_data import get_mask_from_prediction, plot_to_image, get_img_mask_pred_batch_figure_with_id
        
        # define a folder for prediction logging
        file_writer_preds = tf_summary.create_file_writer(log_dir + '/predictions')
        
        # Generate another round of test batches.
        val_batches = test_batches_from_gtiff_dirs_with_names(
                                                    val_img_dir, 
                                                    val_mask_dir, 
                                                    batch_size, 
                                                    input_img_shape, 
                                                    output_img_shape)

        # take a single batch from the data generator.
        # Note it has the same batch_size, height, and width, as the original image
        batch = next(val_batches.take(1).as_numpy_iterator())
        image_batch_tensor = batch[0][0]
        list_of_chip_names = batch[0][1]
        mask_batch_tensor = batch[1][0]
        prediction = model.predict(image_batch_tensor)

        # make a mask from the prediction.
        predmask_batch_tensor = get_mask_from_prediction(prediction)

        # display the result of training.
        img_to_log = plot_to_image(
            get_img_mask_pred_batch_figure_with_id(
                list_of_chip_names, 
                image_batch_tensor, 
                mask_batch_tensor, 
                predmask_batch_tensor)
                )

        # Log the confusion matrix as an image summary.
        with file_writer_preds.as_default():
            tf_summary.image("Test Batch Predictions", img_to_log, step=epoch)
            
    return log_predictions
    
def get_pred_logging_callback_fn(model, cfg):
    return keras.callbacks.LambdaCallback(
        on_epoch_end=get_prediction_logging_fn(model, cfg)
    )

def get_clr_callback_fn(cfg):
    n_training = cfg["runtime"]["n_training"]
    batch_size = cfg["batch_size"]
    
    # get clr parameters
    parms = cfg["cyclic_learning_scheduler"]["clr_callback_parms"]
    mode = parms["mode"]
    step_size = parms["stepsize"]*(n_training//batch_size)
    max_lr = parms["max_lr"]
    base_lr = parms["base_lr"]
    
    # plot clr schedule
    plot_dir = Path(RAMP_HOME) / cfg["cyclic_learning_scheduler"]["clr_plot_dir"] / cfg["timestamp"]
    plot_dir.mkdir(parents=True, exist_ok=True)
    plot_png = str( plot_dir / "clr_schedule_plot.png")
    return CyclicLR(mode=mode, base_lr=base_lr, max_lr = max_lr, step_size = step_size, plot_path= plot_png)

def get_early_stopping_callback_fn(cfg):
    estop_parms = cfg["early_stopping"]["early_stopping_parms"]
    return tf.keras.callbacks.EarlyStopping(**estop_parms)


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import tensorflow as tf

def get_adam_optimizer(cfg):
    adam_parms = cfg["optimizer"]["optimizer_fn_parms"]
    return tf.keras.optimizers.Adam(**adam_parms)


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import numpy as np
from osgeo import gdal, gdalnumeric as gdn, gdalconst as gdc
from skimage.transform import resize
from .ramp_exceptions import GdalReadError

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def to_channels_first(tensor_3):
    '''
    converts a channels-last to a channels-first shape.
    (i.e. from gdal to rasterio)
    '''
    return tensor_3.transpose(2,0,1)

def to_channels_last(tensor_3):
    '''
    converts a channels-first to a channels-last shape
    (from rasterio to gdal)
    '''
    return tensor_3.transpose(1,2,0)

### Detect whether an image has nodata values.

def nodata_count(the_image):
    '''
    image: Float32 np 3-tensor: assumed to be channels-last.
    returns true if the image has nodata values (ie, 0-0-0 at a pixel location)
    '''
    flat_image = the_image.transpose(2,0,1).reshape(3,-1)
    image_sz = flat_image.shape[1]
    inds = np.where((flat_image[0,:] == 0) & (flat_image[1,:] == 0) & (flat_image[2,:] == 0))
    return len(inds[0])


# Standardize and normalize

def standardize_rgb_tensor(the_tensor):
    '''
    rescale the input tensor to have (min, max) = (0,1).
    '''
    ftensor = the_tensor.astype(np.float32)
    the_max = np.max(ftensor)
    the_min = np.min(ftensor)
    span = the_max - the_min
    if span != 0.0:
        ftensor = (ftensor - the_min)/span
    else:
        ftensor = np.zeros(ftensor.shape)
    return ftensor


def normalize_std_tensor(std_tensor):
  '''
  Apply Imagenet normalization to a standardized (0-1) float tensor.  
  '''
  # check that the tensor has 3 channels
  assert std_tensor.shape[-1] == 3, f'unexpected tensor shape: {std_tensor.shape}'

  # apply normalization to each channel
  imgnet_mean = [0.485, 0.456, 0.406] 
  imgnet_stdev = [0.229, 0.224, 0.225]
  norm_tensor = np.zeros(std_tensor.shape)
  for ii in range(3):
    norm_tensor[...,ii] = (std_tensor[...,ii] - imgnet_mean[ii])/imgnet_stdev[ii]
  return norm_tensor


def convert_RGB_array_to_grayscale(rgb_array):
    '''
    converts a channels-last RGB array to a grayscale image using a widely accepted formula.
    :param np.ndarray rgb_array: channels-last RGB array as returned by gdal_get_image_tensor.  
    '''
    R = rgb_array[:,:,0]
    G = rgb_array[:,:,1]
    B = rgb_array[:,:,2]

    # standard conversion formula. Note the coefficients sum to 1.
    return 0.2989 * R  + 0.5870 * G + 0.1140 * B



def dhash(image, hashSize=8):
    '''
    dhash calculates a simple hash function of a channels-last RGB image.
    It's used for finding duplicates in large image datasets.
    The larger hashSize is, the larger the space of possible hashes 
    and the lower is the likelihood of 'hash-collisions' - 
    images that are different but have the same hash. 
    But keep in mind that the space of possible hash values for hashSize=8 is huge (2^64). 
    This is big enough for any reasonable purpose.

    :param np.ndarray image: channels-last RGB image
    '''
	
    # raise NotImplementedError("call to resize is failing in tensorflow, needs fixing")

    # convert the image to grayscale and resize the grayscale image,
	  # adding a single column (width).
    # Note that we are using tensorflow.image.resize(), which requires us to add 
    # a trivial 3rd dimension (using np.newaxis) to the grayscale image before passing it.
    gray = convert_RGB_array_to_grayscale(image)
    # gray_3band = gray[:,:,np.newaxis]
    resized = resize(gray, (hashSize + 1, hashSize), order=0)


	# compute the (relative) horizontal gradient between adjacent
	# column pixels. The result is an array of 0s and 1s of size (hashsize, hashsize)
    diff = resized[:, 1:] > resized[:, :-1]

	# convert the difference image to a hash (integer) and return it
    return sum([2 ** i for (i, v) in enumerate(diff.flatten()) if v])


############################################################
# reading and writing raster geotiff files 
############################################################

def gdal_get_dataset(filename):
  '''
  Open gdal-readable file, return dataset
  '''
  ds  = gdal.Open(filename)
  if not ds:
      log.error(f"GDAL could not read {filename}")
      raise GdalReadError(f"Could not read {filename}")
  return ds


def gdal_get_mask_tensor(mask_path):
    '''
    Reads a mask geotiff image using gdal, and returns a numpy tensor containing the mask.
    Masks have datatype uint8, and a single channel.
    mask_path: path to a gdal-readable image.
    
    Channels go last in the index order for the output tensor, as is standard for tensorflow.

    :param str mask_path: the path to the mask file
    :param str its_dtype: datatype of numpy mask tensor to create. If None, will be inferred from the GDAL datatype.

    returns: a numpy ndarray.
    '''

    ds  = gdal.Open(mask_path)
    if not ds:
      log.error(f"GDAL could not read {mask_path}")
      raise GdalReadError(f"Could not read {mask_path}")

    mask_tensor = get_mask_tensor_from_gdal_dataset(ds)
    return mask_tensor



def gdal_get_image_tensor(image_path):    
  '''
  Reads a satellite image geotiff using gdal, and converts to a numpy tensor.
  Image geotiffs have 3 channels (RGB) and are of type float32.
  image_path: path to a gdal-readable color image.
  
  Channels go last in the index order for the output numpy array.

  :param str image_path: the path to the image file. 
  '''

  its_dtype = "float32"
  ds  = gdal.Open(image_path)
  if not ds:
    log.error(f"GDAL could not read {image_path}")
    raise GdalReadError(f"Could not read {image_path}")
  bands = [ds.GetRasterBand(i) for i in range(1, ds.RasterCount + 1)]
  cols = ds.RasterXSize 
  rows = ds.RasterYSize 
    
  # create a channels_last array -- rows, columns
  the_array = np.zeros((rows, cols, ds.RasterCount), its_dtype)
  
  for bnum, the_band in enumerate(bands):
    from_img = gdn.BandReadAsArray(the_band)
    to_img = the_array[:,:,bnum]
    np.copyto(to_img, from_img)
    
  # normalize the float image to [0,1]
  return the_array/np.max(the_array)

# 2022.04.29 added requirement to specify gdal data type
def gdal_write_mask_tensor(mask_path, src_ds, mask_tensor, gdal_dtype):
  '''
  
  Writes a mask tensor to a new file, with georeferencing matching the gdal dataset.
  The data type of the file to be written depends on the datatype of the mask tensor, which may be either byte or int32.

  :param filename mask_path: path to mask to be created
  :param GdalDataset src_ds: gdal dataset with georeferencing for the mask tensor to match.
  :mask_tensor numpy.ndarray: numpy array containing mask values. Its size should match src_ds or an AssertionError is thrown.
  :gdal_dtype gdal datatype: data type of gdal dataset, usually one of gdal.GDT_Byte, gdal.GDT_Int32, gdal.GDT_Float32
  '''

 # this should create the dataset, then flush it and write out the mask file. 
  mask_ds = create_gdal_dataset_from_mask_tensor(src_ds, mask_tensor, 
                                                  in_ram=False, 
                                                  mask_path=mask_path)
  mask_ds = None
  return


############################################################
# working with GDAL datasets
############################################################

def get_mask_tensor_from_gdal_dataset(ds):
  '''
  extract a tensor from a 1-channel gdal raster dataset.
  :param ds GdalDataset: raster dataset containing the mask

  returns: a numpy ndarray of uint8 datatype
  '''
  assert ds is not None, 'null dataset'

  bands = [ds.GetRasterBand(i) for i in range(1, ds.RasterCount + 1)]
    
  # create a channels_last byte array -- rows, columns
  its_dtype=np.byte
  cols = ds.RasterXSize 
  rows = ds.RasterYSize 
  the_array = np.zeros((rows, cols, ds.RasterCount), its_dtype)
  
  for bnum, the_band in enumerate(bands):
    from_img = gdn.BandReadAsArray(the_band)
    to_img = the_array[:,:,bnum]
    np.copyto(to_img, from_img)
  return the_array


# 2022.04.29 added requirement to specify gdal data type
# 2022.05.20: removed gdal data type requirement (and gdal_dtype input parameter): all masks are byte
def create_gdal_dataset_from_mask_tensor(src_ds, mask_tensor, in_ram=True, mask_path=None):
  '''
  Creates an in-ram or geotiff mask GdalDataset using the mask tensor, with georeferencing matching the gdal dataset.
  In-ram is the default. If in_ram=False is passed, the mask_path must be given.

  :param GdalDataset src_ds: gdal dataset with georeferencing for the mask tensor to match.
  :param filename mask_path: path to mask to be created
  :param mask_tensor numpy.ndarray: numpy array containing mask values. Its size should match src_ds or an AssertionError is thrown.
  :param gdal datatype gdal_dtype: usually one of gdal.GDT_Byte, gdal.GDT_Int32, gdal.GDTFloat32
  Returns the gdal dataset. 
  To flush a geotiff mask to file, set it to None in the calling code. (rolleyes) 
  '''

  gdal_dtype = gdal.GDT_Byte

  # check mask shape 
  mask_shape = mask_tensor.shape
  mask_dims = len(mask_shape)
  sz_x, sz_y = src_ds.RasterXSize, src_ds.RasterYSize
  assert mask_dims == 2 or (mask_dims == 3 and mask_shape[2] == 1), 'mask is neither a 2-tensor nor a 3-tensor with a single channel'

  # if mask is a 3-tensor with a single channel, change it to a 2-tensor (22.04.29: using np.squeeze)
  # No op if it's a 2-tensor
  mask_tensor = np.squeeze(mask_tensor) 
 
  # check reference file dimensions vs. mask tensor dimensions
  # these are reversed in the create functions for gdal vs numpy
  assert mask_tensor.shape[0] == sz_y and mask_tensor.shape[1] == sz_x, \
  'reference image and mask to be created have different shapes'

  # Create a mask raster in memory, matching the parameters of the geotiff image (src_ds).
  # Note: when creating gdal images, the order of dimensions is sz_x, then sz_y
  sz_x, sz_y = src_ds.RasterXSize, src_ds.RasterYSize
  driver = None
  mask_ds = None
  if in_ram:
    driver = gdal.GetDriverByName('MEM')
    mask_ds = driver.Create('',sz_x, sz_y, 1, gdal_dtype)
  else:
    assert mask_path is not None, 'mask path was not given'
    driver = gdal.GetDriverByName('GTiff')
    mask_ds = driver.Create(mask_path,sz_x, sz_y, 1, gdal_dtype)
  assert mask_ds is not None, f'Mask file {mask_path} could not be created'

  mask_ds.SetGeoTransform(src_ds.GetGeoTransform())
  mask_ds.SetProjection(src_ds.GetProjection())
  
  # insert mask array into gdal ds
  mask_ds.GetRasterBand(1).WriteArray(mask_tensor)
  
  # I think those are all the magic incantations you need. 
  return mask_ds


############################################################
# get binary mask, given vector layer and reference GDAL Dataset
############################################################

def get_binary_mask_from_layer(layer, src_ds, in_ram=True, mask_path=None):
  '''
  given a polygon layer and a reference GDAL dataset, 
  create a new binary byte mask and burn the raster into it.
  By default, the mask will be a gdal dataset in ram.
  '''
  # Create a byte mask GDAL dataset, matching the parameters of the geotiff image (src_ds).

  sz_x, sz_y = src_ds.RasterXSize, src_ds.RasterYSize
  blank_array = np.zeros((sz_y, sz_x), dtype = np.byte)
  mask_ds = create_gdal_dataset_from_mask_tensor(src_ds, blank_array, in_ram, mask_path)
  
  # Rasterize the feature into the in-ram mask  
  band = mask_ds.GetRasterBand(1)
  gdal.RasterizeLayer(mask_ds, [1], layer, burn_values=[1])
  band.FlushCache()
  
  # call ReadAsArray on the binary mask raster to get it as a np array.
  mask_array = mask_ds.ReadAsArray()

  # free the mask dataset, or flush the mask to a file if it's not in-ram
  mask_ds = None
  return mask_array
  

############################################################
# get nodata mask from an image
############################################################

def get_nodata_mask_from_image(image_tensor):
  '''
  image_tensor: RGB (float) numpy array of shape (width, height, 3)
  returns mask_tensor: byte numpy array of shape (width, height, 1)
  '''
  mask_tensor_0 = image_tensor[:,:,0]==0.0  
  mask_tensor_1 = image_tensor[:,:,1]==0.0  
  mask_tensor_2 = image_tensor[:,:,2]==0.0

  log.info(f"nodata mask shape: {mask_tensor_2.shape}")
  return_mask = np.logical_and(mask_tensor_0, mask_tensor_1, mask_tensor_2).astype(np.byte)
  return return_mask
  


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import geopandas as gpd
from shapely.geometry import box

def add_overlap_validpoly_columns(gdf):
    '''
    Add both overlap and valid polygon columns to a geodataframe. 
    This function only detects overlaps between valid polygons.
    '''
    # create temporary geodataframe so I can add an index 't_index' column to it
    # this way I have an index column with a known name
    tdf = gdf.copy(deep=True)
    tdf['t_index'] = range(len(tdf))

    # add the valid polygon column to the dataframe
    validpoly_array = [geom.is_valid if geom is not None else 1 for geom in gdf['geometry']]
    tdf['is_valid'] = validpoly_array

    # filter invalid polygons out before checking for overlaps
    valid_tdf = tdf[tdf['is_valid'] == 1]
    
    # do an inner self-join (using 'overlaps' instead of 'intersects').
    joindf= valid_tdf.sjoin(valid_tdf, how="inner", predicate="overlaps")

    # create a list the length of the dataframe that is 1 for every geometry that overlaps another, and 0 otherwise. 
    olap_ids = list(joindf["t_index_left"])
    olap_array = [1 if a in olap_ids else 0 for a in range(len(tdf))]

    # add the new columns to the dataframe
    gdf['is_valid'] = validpoly_array
    gdf["overlaps"] = olap_array
    
    return gdf

def add_overlaps_column(gdf):
    '''
    Add a column to a geodataframe that is 1 if the polygon intersects another polygon, and 0 otherwise.
    '''
    # create temporary geodataframe so I can add an index 't_index' column to it
    # this way I have an index column with a known name

    tdf = gdf.copy(deep=True)
    tdf['t_index'] = range(len(tdf))

    # do an inner self-join on 'overlaps'.
    joindf= tdf.sjoin(tdf, how="inner", predicate="overlaps")

    # create a list the length of the dataframe that is 1 for every geometry that overlaps another, and 0 otherwise. 
    olap_ids = list(joindf["t_index_left"])
    olap_array = [1 if a in olap_ids else 0 for a in range(len(tdf))]

    # add it to the dataframe, return
    gdf["overlaps"] = olap_array
    return gdf


def add_valid_polygon_column(gdf):
    '''
    Add a column to a geodataframe that is 1 if the polygon is valid.
    '''
    validpoly_array = [geom.is_valid for geom in gdf['geometry']]
    gdf['is_valid'] = validpoly_array
    return gdf

def create_tile_gdf (
            raster_tiler, 
            tile_bounds_crs
            ):
    '''
    Create a geojson record of tiles created by the solaris raster and vector tiler.
    '''
    lefts = [tb[0] for tb in raster_tiler.tile_bounds]
    bottoms = [tb[1] for tb in raster_tiler.tile_bounds]
    rights = [tb[2] for tb in raster_tiler.tile_bounds]
    tops = [tb[3] for tb in raster_tiler.tile_bounds]

    tile_records = zip(raster_tiler.tile_paths, lefts, bottoms, rights, tops)

    # construct attributes
    df = gpd.GeoDataFrame(tile_records, columns = ['filename','left', 'bottom', 'right', 'top'])

    # construct geometries
    b = [box(l, b, r, t) for l, b, r, t in zip(df.left, df.bottom, df.right, df.top)]
    gdf = gpd.GeoDataFrame(df, geometry=b)
    gdf = gdf.set_crs(tile_bounds_crs)
    return gdf


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon, MultiPolygon
from shapely.geometry.polygon import orient

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def fill_rings(the_polygon):
    '''
    Returns the polygon with interiors removed.
    '''
    if isinstance(the_polygon, MultiPolygon):
        return the_polygon
    if the_polygon.interiors:
        return Polygon(list(the_polygon.exterior.coords))
    else:
        return the_polygon

def gpd_add_area_perim(geodf):
    '''
    given a geodataframe containing a column named 'geometry' with polygons, 
    add an area and a perimeter field to it, and return it. 
    :param geopandas.DataFrame geodf: a geodataframe with a polygon field named 'geometry'.
    '''
    if not (geodf.crs and geodf.crs.is_geographic):
        log.error('geodataframe does not have a geographic coordinate system')
        raise TypeError('the geodataframe should have a geographic coordinate system')
        
    geod = geodf.crs.get_geod()

    # Note: I'm defining these functions internally to gpd_add_area_perim, 
    # so they have access to the geod object.

    def area_calc(geom):
        if geom.geom_type not in ['MultiPolygon','Polygon']:
            log.warning(f"geometry type is {geom.geom_type} -- must be one of MultiPolygon, Polygon")
            return np.nan
        
        # For MultiPolygon do each separately
        if geom.geom_type=='MultiPolygon':
            return np.sum([area_calc(p) for p in geom.geoms])

        # orient to ensure a counter-clockwise traversal. 
        # See https://pyproj4.github.io/pyproj/stable/api/geod.html
        # geometry_area_perimeter returns (area, perimeter)
        return geod.geometry_area_perimeter(orient(geom, 1))[0]

    def perim_calc(geom):
        if geom.geom_type not in ['MultiPolygon','Polygon']:
            log.warning(f"geometry type is {geom.geom_type} -- must be one of MultiPolygon, Polygon")
            return np.nan
        
        # For MultiPolygon do each separately
        if geom.geom_type=='MultiPolygon':
            return np.sum([perim_calc(p) for p in geom.geoms])

        # orient to ensure a counter-clockwise traversal. 
        # See https://pyproj4.github.io/pyproj/stable/api/geod.html
        # geometry_area_perimeter returns (area, perimeter)
        return geod.geometry_area_perimeter(orient(geom, 1))[1]

    
    area_series = geodf.geometry.apply(area_calc)
    perim_series = geodf.geometry.apply(perim_calc)
    geodf['area_m2'] = area_series
    geodf['perim_m'] = perim_series
    return geodf


def get_polygon_indices_to_merge(df, pairs_df):
    '''
    used in 'get_labels_from_masks.py' to identify polygons that should be merged 
    from the combined outputs of chipwise polygonization.

    df: geopandas.GeoDataFrame: containing polygons to be merged
    pairs_df: geopandas.GeoDataFrame: containing the result of a spatial join between df and buffered df,
        and postfiltered to retain only boundary-crossing pairs of polygons
    '''

    equivalence_classes = []
    for i1, row1 in pairs_df.iterrows():
        found_class=False
        polyid_left = row1["polyid_left"]
        polyid_right = row1["polyid_right"]

        # if polyid_left == 324 or polyid_right== 324:
        #     print("stop")

        # loop over equivalence classes
        for a_class in equivalence_classes:
            if polyid_left in a_class or polyid_right in a_class:
                a_class.update([polyid_right, polyid_left])
                found_class = True
                break # out of equivalence class loop
 
        # if found_class is false, you've discovered a new equivalence class of polygons to merge
        if not found_class:
            new_class = set([polyid_left, polyid_right])
            equivalence_classes.append(new_class)

    # Create the 'merge_class' attribute 
    # every polygon is in its own equivalence class to start with
    eq_class_list = df["polyid"].copy()

    # start defining new equivalence class indices greater than polyid values
    for ii, eq_class in enumerate(equivalence_classes):
        current_eq_class_index = len(eq_class_list) + ii

        # set all polyids in the same equivalence class to the same merge class
        for polyid in eq_class:
            eq_class_list[polyid] = current_eq_class_index
    
    df["merge_class"]= eq_class_list
    return df
                    



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import os
import io
import numpy as np
import matplotlib.pyplot as plt
from tensorflow.image import decode_png
from tensorflow import expand_dims
import tensorflow.summary as tf_summary

from ..utils.img_utils import gdal_get_image_tensor, gdal_get_mask_tensor
from ..utils.file_utils import get_basename
from .data_generator import test_batches_from_gtiff_dirs

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


# Taken from a tensorboard image logging demo at: https://www.tensorflow.org/tensorboard/image_summaries
def plot_to_image(figure):
  """
  Converts the matplotlib plot specified by 'figure' to a PNG image and
  returns it. 
  The supplied figure is closed and inaccessible after this call.
  This code is from a tensorboard/tensorflow image logging demo at: https://www.tensorflow.org/tensorboard/image_summaries
  """
  # Save the plot to a PNG in memory.
  buf = io.BytesIO()
  plt.savefig(buf, format='png')
  # Closing the figure prevents it from being displayed directly inside
  # the notebook.
  plt.close(figure)
  buf.seek(0)
  # Convert PNG buffer to an RGBA (RGB - alpha) TF image to be logged
  image = decode_png(buf.getvalue(), channels=4)
  # Add the batch dimension
  image = expand_dims(image, 0)
  return image


def get_mask_from_prediction(prediction):
    """
    Quick utility to display a model's prediction on a test batch.
    """
    predmask = np.argmax(prediction, axis=-1)
    predmask = np.expand_dims(predmask, axis=-1)
    return predmask

def get_offset_mask_from_prediction(prediction, mask_offset=0):
    """
    Utility to obtain multichannel or binary mask from a one-hot-encoded prediction
    """
    predmask = np.argmax(prediction, axis=-1)
    predmask = np.expand_dims(predmask, axis=-1)
    return predmask + mask_offset

###########################################################################################

def display_a_batch(a_batch):
    '''
    a_batch: a 2-tuple of the form (image_batch, mask_batch)
    image_batch has shape: (batch_size, height, width, 3)
    mask_batch has shape: (batch_size, height, width, 1)
    displays image data in a jupyter notebook 
    '''

    # zip the data together as:
    # (image-mask_1|image-mask_2|...|image-mask_B)
    # where B is batch size
    datapts = list(zip(a_batch[0], a_batch[1]))
    B = len(datapts)

    plt.figure(figsize=(15,(0.5*B*15)))
    
    title = ['Image', 'Mask']
    
    for ii, dd in enumerate(datapts):        
        for jj in range(2):
            plt.subplot(B, 2, ii*2 + jj +1)
            plt.title(title[jj])
            plt.imshow(dd[jj])
    plt.axis('off')
    plt.show()

    ###########################################################################################

# added 20220713

def get_img_mask_pred_batch_figure_with_id(b_idstr, b_img, b_mask, b_predmask):
    '''
    b_idstr: batch id string (list of chips)
    image batch has shape: (batch_size, height, width, 3)
    mask batch and predmask batch both have shape: (batch_size, height, width, 1).
    Run 'get_mask_from_prediction' on the output of model.predict to get b_predmask before calling this function.
    It displays image data in a jupyter notebook.
    '''

    # extract the tile ids from the list of filepaths.
    tile_ids = [get_basename(fpath) for fpath in b_idstr]

    # zip the data together as:
    # (id-image-mask-pred_1|id-image-mask-pred_2|...|id-image-mask-pred_B)
    # where B is batch size
    datapts = list(zip(tile_ids, b_img, b_mask, b_predmask))
    B = len(datapts) # batch size

    figure = plt.figure(figsize=(15,(0.5*B*15)))

    

    for ii, dd in enumerate(datapts):

        # implant chip name in title
        title = [f'Image ({dd[0]})', 'Mask', "Predicted"]
        for jj in range(1,4):
            plt.subplot(B, 3, ii*3 + jj)
            plt.title(title[jj-1])
            plt.imshow(dd[jj])
    plt.axis('off')
    return figure

###################################################################    

def get_img_mask_pred_batch_figure(b_img, b_mask, b_predmask):
    '''
    image batch has shape: (batch_size, height, width, 3)
    mask batch and predmask batch both have shape: (batch_size, height, width, 1).
    Run 'get_mask_from_prediction' on the output of model.predict to get b_predmask before calling this function.
    It displays image data in a jupyter notebook.
    '''
    # zip the data together as:
    # (image-mask-pred_1|image-mask-pred_2|...|image-mask-pred_B)
    # where B is batch size
    datapts = list(zip(b_img, b_mask, b_predmask))
    B = len(datapts) # batch size

    figure = plt.figure(figsize=(15,(0.5*B*15)))

    title = ['Image', 'Mask', "Predicted"]

    for ii, dd in enumerate(datapts):
        for jj in range(3):
            plt.subplot(B, 3, ii*3 + jj +1)
            plt.title(title[jj])
            plt.imshow(dd[jj])
    plt.axis('off')
    return figure


def display_img_mask_pred_batch(b_img, b_mask, b_predmask):
    figure = get_img_mask_pred_batch_figure(b_img, b_mask, b_predmask)
    plt.show()


def display_image_mask(image_path, mask_path):
    '''
    Display the first image passed as an RGB image, 
    and the second as a grayscale mask.
    The user's in charge of making sure they match.
    '''

    img_tensor = gdal_get_image_tensor(image_path)
    mask_tensor = gdal_get_mask_tensor(mask_path)
    trivial_batch = (img_tensor[np.newaxis, :,:,:], mask_tensor[np.newaxis, :,:,:])
    display_a_batch(trivial_batch)
    return

def get_prediction_logging_function(log_dir, model, test_img_dir, test_mask_dir, batch_size, input_image_size, output_image_size):
    '''
    Defines and return a callback function that will log predicted segmentation images 
    from a fixed test data batch after every epoch, during training.
    :param str log_dir: directory to log images to (in the /predictions folder)
    :param str model: model being fitted
    :param str test_img_dir: path to directory with test images
    :param str test_mask_dir: path to directory with test masks
    :param int batch_size: number of images from the beginning of the test image list to use in logging
    :param tuple input_image_size: 2-tuple of integers (width and height of model input images and masks)
    :param tuple output_image_size: 2-tuple of integers (width and height of model output images and masks)
    '''
    log.debug(f"logging predicted masks to logging directory: {log_dir}")
    log.debug(f"model class: {type(model)}")
    log.debug(f"test image directory: {test_img_dir}")
    log.debug(f"test mask directory: {test_mask_dir}")
    log.debug(f"batch_size: {batch_size}")
    log.debug(f"input image size: {input_image_size}")
    log.debug(f"output image size: {output_image_size}")

    def log_predictions(epoch, logs):
        '''
        callback function for logging predictions after every epoch.
        '''

        # define a folder for prediction logging
        file_writer_preds = tf_summary.create_file_writer(log_dir + '/predictions')
        
        # Generate another round of test batches.
        val_batches = test_batches_from_gtiff_dirs(test_img_dir, 
                                                     test_mask_dir, 
                                                     batch_size, 
                                                     input_image_size, 
                                                     output_image_size)

        # take a single batch from the data generator.
        # Note it has the same batch_size, height, and width, as the original image
        # Since this is a prediction problem with 2 classes, the prediction has 2 channels. 
        batch = next(val_batches.take(1).as_numpy_iterator())
        prediction = model.predict(batch[0])

        # make a mask from the prediction.
        predmask = get_mask_from_prediction(prediction)

        # display the result of training.
        img_to_log = plot_to_image(get_img_mask_pred_batch_figure(batch[0], batch[1], predmask))

        # Log the confusion matrix as an image summary.
        with file_writer_preds.as_default():
            tf_summary.image("Test Batch Predictions", img_to_log, step=epoch)
            
    return log_predictions



#################################################################
#
# created for ramp project, August 2022
#
#################################################################


### Forked from solaris sol.vector.mask on 20220701

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

from solaris.utils.core import _check_df_load, _check_gdf_load, _check_geom, _check_crs
from solaris.utils.core import _check_skimage_im_load, _check_rasterio_im_load
from solaris.utils.geo import gdf_get_projection_unit, reproject
from solaris.utils.geo import geometries_internal_intersection
from solaris.utils.tile import save_empty_geojson

# note that affine_transform_gdf from this import will be overwritten by the local version.
# replace with import by name!
from solaris.vector.polygon import convert_poly_coords, geojson_to_px_gdf
import numpy as np
from shapely.geometry import shape
from shapely.geometry import Polygon
import geopandas as gpd
import pandas as pd
import rasterio
from rasterio import features
from affine import Affine
from skimage.morphology import square, erosion, dilation
import os
from tqdm.auto import tqdm

def df_to_px_mask(
                df, 
                channels=['footprint'], 
                out_file=None, 
                reference_im=None,
                geom_col='geometry', 
                do_transform=None, 
                affine_obj=None,
                shape=(900, 900), 
                out_type='int', 
                burn_value=255, 
                **kwargs
                ):
    """Convert a dataframe of geometries to a pixel mask.

    Arguments
    ---------
    df : :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame`
        A :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame` instance
        with a column containing geometries (identified by `geom_col`). If the
        geometries in `df` are not in pixel coordinates, then `affine` or
        `reference_im` must be passed to provide the transformation to convert.
    channels : list, optional
        The mask channels to generate. There are three values that this can
        contain:

        - ``"footprint"``: Create a full footprint mask, with 0s at pixels
            that don't fall within geometries and `burn_value` at pixels that
            do.
        - ``"boundary"``: Create a mask with geometries outlined. Use
            `boundary_width` to set how thick the boundary will be drawn.
        - ``"contact"``: Create a mask with regions between >= 2 closely
            juxtaposed geometries labeled. Use `contact_spacing` to set the
            maximum spacing between polygons to be labeled.

        Each channel correspond to its own `shape` plane in the output.
    out_file : str, optional
        Path to an image file to save the output to. Must be compatible with
        :class:`rasterio.DatasetReader`. If provided, a `reference_im` must be
        provided (for metadata purposes).
    reference_im : :class:`rasterio.DatasetReader` or `str`, optional
        An image to extract necessary coordinate information from: the
        affine transformation matrix, the image extent, etc. If provided,
        `affine_obj` and `shape` are ignored.
    geom_col : str, optional
        The column containing geometries in `df`. Defaults to ``"geometry"``.
    do_transform : bool, optional
        Should the values in `df` be transformed from geospatial coordinates
        to pixel coordinates? Defaults to ``None``, in which case the function
        attempts to infer whether or not a transformation is required based on
        the presence or absence of a CRS in `df`. If ``True``, either
        `reference_im` or `affine_obj` must be provided as a source for the
        the required affine transformation matrix.
    affine_obj : `list` or :class:`affine.Affine`, optional
        Affine transformation to use to convert from geo coordinates to pixel
        space. Only provide this argument if `df` is a
        :class:`geopandas.GeoDataFrame` with coordinates in a georeferenced
        coordinate space. Ignored if `reference_im` is provided.
    shape : tuple, optional
        An ``(x_size, y_size)`` tuple defining the pixel extent of the output
        mask. Ignored if `reference_im` is provided.
    burn_value : `int` or `float`
        The value to use for labeling objects in the mask. Defaults to 255 (the
        max value for ``uint8`` arrays). The mask array will be set to the same
        dtype as `burn_value`.
    kwargs
        Additional arguments to pass to `boundary_mask` or `contact_mask`. See
        those functions for requirements.

    Returns
    -------
    mask : :class:`numpy.array`
        A pixel mask with 0s for non-object pixels and `burn_value` at object
        pixels. `mask` dtype will coincide with `burn_value`. Shape will be
        ``(shape[0], shape[1], len(channels))``, with channels ordered per the
        provided `channels` `list`.

    """
    if isinstance(channels, str):  # e.g. if "contact", not ["contact"]
        channels = [channels]

    if out_file and not reference_im:
        raise ValueError(
            'If saving output to file, `reference_im` must be provided.')

    mask_dict = {}

    meters = kwargs.get('meters',False)

    # 20220802 bugfix: clean up geometries before buffer_df_geoms is called
    df = buffer_df_geoms(df, 0, meters=meters, reference_im=reference_im)


    if 'footprint' in channels:
        mask_dict['footprint'] = shapes_mask(
            df=df, 
            reference_im=reference_im, 
            geom_col=geom_col,
            do_transform=do_transform, 
            shape=shape,
            burn_value=burn_value,
            out_file=None
        )
    if 'boundary' in channels:

        try:
            interior_df = get_interior_dataframe(
                            df, 
                            geom_col=geom_col,
                            reference_im=reference_im,
                            meters=kwargs.get('meters',False),
                            boundary_width=kwargs.get('boundary_width', 3),
                            out_file=None
            )
                            # out_file="/tf/ramp-data/temp.file.geojson")

            interior_mask = shapes_mask(
                            df=interior_df, 
                            reference_im=reference_im, 
                            geom_col=geom_col,
                            do_transform=do_transform, 
                            shape=shape,
                            burn_value=burn_value,
                            out_file=None
                            )
                            # out_file="/tf/ramp-data/temp.file.tif")

            mask_dict['boundary'] = mask_dict['footprint'] - interior_mask

        # handle these key errors that arise when there is a problem with negative buffering
        except KeyError:
            log.info(f"could not determine boundary pixels in df_to_px_mask (got KeyError), bypassing")
            mask_dict["boundary"] = np.zeros(shape)

    if 'contact' in channels:

        try:
            contact_df = get_contact_dataframe(
                            df,
                            geom_col=geom_col,
                            reference_im=reference_im,
                            meters=kwargs.get('meters', False),
                            contact_spacing=kwargs.get('contact_spacing', 8)
                            )


            mask_dict['contact'] = shapes_mask(
                            df=contact_df, 
                            reference_im=reference_im, 
                            geom_col=geom_col,
                            do_transform=do_transform, 
                            shape=shape,
                            burn_value=burn_value,
                            out_file=None)

        except KeyError:
            log.info(f"could not determine close contact pixels in df_to_px_mask (got KeyError), bypassing")
            mask_dict["contact"] = np.zeros(shape)
            
    output_arr = np.stack([mask_dict[c] for c in channels], axis=-1)

    if reference_im:
        reference_im = _check_rasterio_im_load(reference_im)
    if out_file:
        meta = reference_im.meta.copy()
        meta.update(count=output_arr.shape[-1])
        meta.update(dtype='uint8')
        with rasterio.open(out_file, 'w', **meta) as dst:
            # I hate band indexing.
            for c in range(1, 1 + output_arr.shape[-1]):
                dst.write(output_arr[:, :, c-1], indexes=c)

    return output_arr

    ##############################################################################

def shapes_mask(df, 
                out_file, 
                reference_im, 
                geom_col,
                do_transform,
                shape,
                burn_value):
    """Convert a dataframe of geometries to a pixel mask.

    Arguments
    ---------
    df : :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame`
        A :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame` instance
        with a column containing geometries (identified by `geom_col`). If the
        geometries in `df` are not in pixel coordinates, then `affine` or
        `reference_im` must be passed to provide the transformation to convert.
    out_file : str, optional
        Path to an image file to save the output to. Must be compatible with
        :class:`rasterio.DatasetReader`. If provided, a `reference_im` must be
        provided (for metadata purposes).
    reference_im : :class:`rasterio.DatasetReader` or `str`, optional
        An image to extract necessary coordinate information from: the
        affine transformation matrix, the image extent, etc. 
    geom_col : str, optional
        The column containing geometries in `df`. Defaults to ``"geometry"``.
    do_transform : bool, optional
        Should the values in `df` be transformed from geospatial coordinates
        to pixel coordinates? 
        Defaults to ``None``, in which case the function
        attempts to infer whether or not a transformation is required based on
        the presence or absence of a CRS in `df`. 
        If ``True``, either
        `reference_im` or `affine_obj` must be provided as a source for the
        the required affine transformation matrix.
    burn_value : `int` or `float`, optional
        The value to use for labeling objects in the mask. Defaults to 255 (the
        max value for ``uint8`` arrays). The mask array will be set to the same
        dtype as `burn_value`. Ignored if `burn_field` is provided.

    Returns
    -------
    mask : :class:`numpy.array`
        A pixel mask with 0s for non-object pixels and `burn_value` at object
        pixels. `mask` dtype will coincide with `burn_value`.

    """
    # start with required checks and pre-population of values

    if len(df) == 0 and not out_file:
        return np.zeros(shape=shape, dtype='uint8')

    affine_obj = None
    if not do_transform:
        affine_obj = Affine(1, 0, 0, 0, 1, 0)  # identity transform

    shape = None
    if reference_im:
        reference_im = _check_rasterio_im_load(reference_im)
        shape = reference_im.shape
        if do_transform:
            affine_obj = reference_im.transform

    # extract geometries 
    feature_list = list(df[geom_col])

    if len(df) > 0:
        output_arr = features.rasterize(shapes=feature_list, 
                                        fill=0,
                                        default_value=burn_value,
                                        out_shape=shape,
                                        transform=affine_obj,
                                        dtype="uint8")
    else:
        output_arr = np.zeros(shape=shape, dtype='uint8')
    
    if out_file:
        meta = reference_im.meta.copy()
        meta.update(count=1)
        meta.update(dtype='uint8')
        meta.update(nodata=None)
        with rasterio.open(out_file, 'w', **meta) as dst:
            output_mask = np.expand_dims(output_arr, axis=0)
            dst.write(output_mask)

    return output_arr

##################################################################################################
## CJ 20220705: convert footprint-bdry-contact mask to sparse mask

def multimask_to_sparse_multimask(onehot_mask):
    """
    Change a one-hot (channels-last) multimask to a sparse-encoded multimask
    bdry > footprint > contact
    onehot_mask: numpy ndarray of shape [H,W, C] with values 0,1
    sparse_mask: numpy ndarray of shape [H,W,1] with values 0,...,C-1 
    """
    sparse_mask = np.zeros(onehot_mask.shape[:2], dtype=np.uint8)
    sparse_mask[np.where(onehot_mask[:,:,2] > 0)] = 3 #contact
    sparse_mask[np.where(onehot_mask[:,:,0] > 0)] = 1 #footprint
    sparse_mask[np.where(onehot_mask[:,:,1] > 0)] = 2 #boundary
    return np.expand_dims(sparse_mask, axis=-1)


## CJ 20220704: get shapes to burn into rasters for boundary and contact points. 
def get_interior_dataframe(
                df, 
                boundary_width, 
                reference_im, 
                meters,
                geom_col="geometry",
                out_file=None
                ):
    '''
    get a dataframe of negative-buffered building masks. 
    The difference between the original mask, and this mask, becomes the building boundary mask. -
    '''
    interior_df = buffer_df_geoms(
                        df, 
                        buffer=-boundary_width, 
                        reference_im=reference_im, 
                        meters=meters
                        )

    # remove empty geometries
    interior_df = interior_df[~interior_df.is_empty]

    # return interior df
    if out_file:
        interior_df.to_file(out_file)
    return interior_df


def get_contact_dataframe(
                    df,
                    contact_spacing, 
                    reference_im,
                    meters, 
                    geom_col
                    ):
    '''
    '''
    df['t_index'] = range(len(df))

    # create a new copy of the df with buffered geometries.
    buffer_df = buffer_df_geoms(
                    df, 
                    buffer=contact_spacing,
                    reference_im=reference_im
                    )

    # do a spatial inner join of buffered geometries with unbuffered geometries.
    joined_df = buffer_df.overlay(df, how="intersection")

    # remove the self-intersections
    joined_df = joined_df[~(joined_df['t_index_1']==joined_df["t_index_2"])]
    
    # remove empty polygons
    joined_df = joined_df[~joined_df.is_empty]

    # If B(X) represents the buffered version of shape X, then
    # this df contains shapes of the form B(X)\intersect Y for all X, Y in df
    # if B(X)\intersect Y is in the df, so is B(Y)\intersect X
    joined_df = joined_df[['t_index_1', 't_index_2', "geometry"]]

    # eject if there are no nontrivial polygons
    if len(joined_df) == 0:
        return joined_df

        
    # next: if B(X)\intersect Y is in the df, then we want to get B(X)\intersect B(Y)
    # we need to index into the data frame of buffered shapes, which is tricky to get right.
    # This function will be applied to rows in joined_df using 'apply'.
    # It returns the polygon in buffer_df pointed to by t_index_1.
    def get_buff_geom_1(x): 
        return buffer_df[buffer_df["t_index"]==x.t_index_1]["geometry"].values[0]

    # This one returns the polygon in buffer_df pointed to by t_index_2.
    def get_buff_geom_2(x): 
        return buffer_df[buffer_df["t_index"]==x.t_index_2]["geometry"].values[0]

    # Now we apply the lookups to get the buffered polygons.
    b1 = joined_df.apply(lambda x: get_buff_geom_1(x), axis=1  )
    b2 = joined_df.apply(lambda x: get_buff_geom_2(x), axis=1)

    # Note: b1 and b2 are pandas series which happen to contain polygons! 
    # they are not GeoSeries so you can't run geo operations on them.
    # here I turn them into GeoSeries so I can call 'intersection' on them
    buff1 = gpd.GeoSeries(b1, crs=joined_df.crs )
    buff2 = gpd.GeoSeries(b2, crs=joined_df.crs)

    # this returns the intersection polygons, which cover the sets of contact points. 
    # Note that since B(X)\intersect B(Y)=B(Y)\intersect B(X),
    # each polygon will appear twice, but we don't mind. 
    buffer_ints = buff1.intersection(buff2, align=True)

    # Construct a geodataframe with the intersection polygons 
    # and the indices in the original dataset. Return it. 
    gdf_data = {
        "t1":joined_df["t_index_1"],
        "t2":joined_df["t_index_2"],
        "geometry":buffer_ints
    }
    contactpts_df = gpd.GeoDataFrame(data=gdf_data, crs=joined_df.crs)
    return contactpts_df

        
##########################################################################################
## ORIGINAL SOLARIS
##########################################################################################


def affine_transform_gdf(gdf, affine_obj, inverse=False, geom_col="geometry",
                         precision=None):
    """Perform an affine transformation on a GeoDataFrame.
    
    Originally from solaris repository but contains a bugfix.

    Arguments
    ---------
    gdf : :class:`geopandas.GeoDataFrame`, :class:`pandas.DataFrame`, or `str`
        A GeoDataFrame, pandas DataFrame with a ``"geometry"`` column (or a
        different column containing geometries, identified by `geom_col` -
        note that this column will be renamed ``"geometry"`` for ease of use
        with geopandas), or the path to a saved file in .geojson or .csv
        format.
    affine_obj : list or :class:`affine.Affine`
        An affine transformation to apply to `geom` in the form of an
        ``[a, b, d, e, xoff, yoff]`` list or an :class:`affine.Affine` object.
    inverse : bool, optional
        Use this argument to perform the inverse transformation.
    geom_col : str, optional
        The column in `gdf` corresponding to the geometry. Defaults to
        ``'geometry'``.
    precision : int, optional
        Decimal precision to round the geometries to. If not provided, no
        rounding is performed.
    """
    if isinstance(gdf, str):  # assume it's a geojson
        if gdf.lower().endswith('json'):
            gdf = gpd.read_file(gdf)
        elif gdf.lower().endswith('csv'):
            gdf = pd.read_csv(gdf)
        else:
            raise ValueError(
                "The file format is incompatible with this function.")
    if 'geometry' not in gdf.columns:
        gdf = gdf.rename(columns={geom_col: 'geometry'})


    # Check datatype of gdf["geometry"]: if it's not a Polygon, 
    # assume its a WKT string and load it.
    
    # CJ bugfix 20220728: indexing into gdf["geometry"] is done with the GeoSeries index,
    # not with the list index! So if list index 0 is not present, a call to 
    # gdf["geometry"][0] will throw a KeyError. 
    # I get around it by retrieving its keys and checking the type of the first one. 
#    if not isinstance(gdf['geometry'][0], Polygon):

    gdf_keys = gdf['geometry'].keys()
    first_key = gdf_keys[0]
    if not isinstance(gdf['geometry'][first_key], Polygon):
        gdf['geometry'] = gdf['geometry'].apply(shapely.wkt.loads)

    gdf["geometry"] = gdf["geometry"].apply(convert_poly_coords,
                                            affine_obj=affine_obj,
                                            inverse=inverse)
    if precision is not None:
        gdf['geometry'] = gdf['geometry'].apply(
            _reduce_geom_precision, precision=precision)

    # the CRS is no longer valid - remove it
    gdf.crs = None

    return gdf


def buffer_df_geoms(df, buffer, meters=False, reference_im=None,
                    geom_col='geometry', affine_obj=None):
    """Buffer geometries within a pd.DataFrame or gpd.GeoDataFrame.

    Arguments
    ---------
    df : :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame`
        A :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame` instance
        with a column containing geometries (identified by `geom_col`). If `df`
        lacks a ``crs`` attribute (isn't a :class:`geopandas.GeoDataFrame` )
        and ``meters=True``, then `reference_im` must be provided for
        georeferencing.
    buffer : `int` or `float`
        The amount to buffer the geometries in `df`. In pixel units unless
        ``meters=True``. This corresponds to width/2 in mask creation
        functions.
    meters : bool, optional
        Should buffers be in pixel units (default) or metric units (if `meters`
        is ``True``)?
    reference_im : `str` or :class:`rasterio.DatasetReader`, optional
        The path to a reference image covering the same geographic extent as
        the area labeled in `df`. Provided for georeferencing of pixel
        coordinate geometries in `df` or conversion of georeferenced geometries
        to pixel coordinates as needed. Required if `meters` is ``True`` and
        `df` lacks a ``crs`` attribute.
    geom_col : str, optional
        The column containing geometries in `df`. Defaults to ``"geometry"``.
    affine_obj : `list` or :class:`affine.Affine`, optional
        Affine transformation to use to convert geoms in `df` from a geographic
        crs to pixel space. Only provide this argument if `df` is a
        :class:`geopandas.GeoDataFrame` with coordinates in a georeferenced
        coordinate space. Ignored if `reference_im` is provided.

    Returns
    -------
    buffered_df : :class:`pandas.DataFrame`
        A :class:`pandas.DataFrame` in the original coordinate reference system
        with objects buffered per `buffer`.

    See Also
    --------
    road_mask : Function to create road network masks.
    contact_mask : Function to create masks of contact points between polygons.
    """

    df = df[~df.is_empty]

    if len(df)==0:
        return df.copy()

    if reference_im is not None:
        reference_im = _check_rasterio_im_load(reference_im)

    if hasattr(df, 'crs'):
        orig_crs = _check_crs(df.crs)
    else:
        orig_crs = None  # will represent pixel crs

    # Check if dataframe is in the appropriate units and reproject if not
    if not meters:
        if hasattr(df, 'crs') and reference_im is not None:
            # if the df is georeferenced and a reference_im is provided,
            # use reference_im to transform df to px coordinates
            df_for_buffer = geojson_to_px_gdf(df.copy(), reference_im)
        elif hasattr(df, 'crs') and reference_im is None:

            # CJ 20220728: bugfix: inverse needs to be used here
            df_for_buffer = affine_transform_gdf(df.copy(),
                                                 affine_obj=affine_obj,
                                                 inverse=True)
        else:  # if it's already in px coordinates
            df_for_buffer = df.copy()

    else:
        # check if the df is in a metric crs
        if hasattr(df, 'crs'):
            if crs_is_metric(df):
                df_for_buffer = df.copy()
            else:
                df_for_buffer = reproject(df.copy())  # defaults to UTM
        else:
            # assume df is in px coords - use reference_im to georegister
            if reference_im is not None:
                df_for_buffer = georegister_px_df(df.copy(),
                                                  im_path=reference_im)
            else:
                raise ValueError('If using `meters=True`, either `df` must be '
                                 'a geopandas GeoDataFrame or `reference_im` '
                                 'must be provided for georegistration.')

    df_for_buffer[geom_col] = df_for_buffer[geom_col].apply(
        lambda x: x.buffer(buffer))

    # sometimes negative buffering can produce multipolygons
    df_for_buffer = df_for_buffer.explode(ignore_index=True) 

    # return to original crs
    if _check_crs(getattr(df_for_buffer, 'crs', None)) != orig_crs:
        if orig_crs is not None and \
                getattr(df_for_buffer, 'crs', None) is not None:
            buffered_df = df_for_buffer.to_crs(orig_crs.to_wkt())
        elif orig_crs is None:  # but df_for_buffer has one: meters=True case
            buffered_df = geojson_to_px_gdf(df_for_buffer, reference_im)
        else:  # orig_crs exists, but df_for_buffer doesn't have one
            buffered_df = georegister_px_df(df_for_buffer,
                                            im_path=reference_im,
                                            affine_obj=affine_obj,
                                            crs=orig_crs)
    else:
        buffered_df = df_for_buffer

    return buffered_df


def preds_to_binary(pred_arr, channel_scaling=None, bg_threshold=0):
    """Convert a set of predictions from a neural net to a binary mask.

    Arguments
    ---------
    pred_arr : :class:`numpy.ndarray`
        A set of predictions generated by a neural net (generally in ``float``
        dtype). This can be a 2D array or a 3D array, in which case it will
        be convered to a 2D mask output with optional channel scaling (see
        the `channel_scaling` argument). If a filename is provided instead of
        an array, the image will be loaded using scikit-image.
    channel_scaling : `list`-like of `float`s, optional
        If `pred_arr` is a 3D array, this argument defines how each channel
        will be combined to generate a binary output. channel_scaling should
        be a `list`-like of length equal to the number of channels in
        `pred_arr`. The following operation will be performed to convert the
        multi-channel prediction to a 2D output ::

            sum(pred_arr[channel]*channel_scaling[channel])

        If not provided, no scaling will be performend and channels will be
        summed.

    bg_threshold : `int` or `float`, optional
        The cutoff to set to distinguish between background and foreground
        pixels in the final binary mask. Binarization takes place *after*
        channel scaling and summation (if applicable). Defaults to 0.

    Returns
    -------
    mask_arr : :class:`numpy.ndarray`
        A 2D boolean ``numpy`` array with ``True`` for foreground pixels and
        ``False`` for background.
    """
    pred_arr = _check_skimage_im_load(pred_arr).copy()

    if len(pred_arr.shape) == 3:
        if pred_arr.shape[0] < pred_arr.shape[-1]:
            pred_arr = np.moveaxis(pred_arr, 0, -1)
        if channel_scaling is None:  # if scale values weren't provided
            channel_scaling = np.ones(shape=(pred_arr.shape[-1]),
                                      dtype='float')
        pred_arr = np.sum(pred_arr*np.array(channel_scaling), axis=-1)

    mask_arr = (pred_arr > bg_threshold).astype('uint8')

    return mask_arr*255

def binary_mask_to_geodataframe(binary_array,
                                affine_object=None,
                                df_crs="EPSG:4326", 
                                do_transform=False,
                                simplify=False,
                                tolerance=0.5,
                                min_area=0,
                                **kwargs):
    """
    Get a polygon geodataframe from a binary mask.
    Simplified version of 'mask_to_poly_geojson' that takes an affine object for transformation
    instead of a reference image.

    Arguments
    ---------
    binary_array : :class:`numpy.ndarray`
        A 2D array of integers with values in 0,1.
    affine_object : object of class Affine 
        for calculating the final geotransform.
        Required if 'do_transform' is True
    df_crs: EPSG string for the CRS of the output geodataframe
        Required if 'do_transform' is true
    min_area : int, optional
        The minimum area of a polygon to retain. Filtering is done AFTER
        any coordinate transformation, and therefore will be in destination
        units.
    simplify : bool, optional
        If ``True``, will use the Douglas-Peucker algorithm to simplify edges,
        saving memory and processing time later. Defaults to ``False``.
    tolerance : float, optional
        The tolerance value to use for simplification with the Douglas-Peucker
        algorithm. Defaults to ``0.5``. Only has an effect if
        ``simplify=True``.

    Returns
    -------
    gdf : :class:`geopandas.GeoDataFrame`
        A GeoDataFrame of polygons.

    """

    if do_transform and affine_object is None:
        raise ValueError(
            'Coordinate transformation requires an input affine transformation.')

    if do_transform:
        transform = affine_object
        crs = _check_crs(df_crs)
    else:
        transform = Affine(1, 0, 0, 0, 1, 0)  # identity transform
        crs = rasterio.crs.CRS()

    # pixels not in 'mask' are excluded from feature generation,
    # so we generate only buildings
    mask = binary_array > 0
    mask = mask.astype('uint8')

    polygon_generator = features.shapes(binary_array,
                                        transform=transform,
                                        mask=mask)
    polygons = []
    values = []  # pixel values for the polygon in mask_arr
    for polygon, value in polygon_generator:
        p = shape(polygon).buffer(0.0)
        if p.area >= min_area:
            polygons.append(shape(polygon).buffer(0.0))
            values.append(value)

    polygon_gdf = gpd.GeoDataFrame({'geometry': polygons, 'value': values},
                                   crs=crs.to_wkt())
    if simplify:
        polygon_gdf['geometry'] = polygon_gdf['geometry'].apply(
            lambda x: x.simplify(tolerance=tolerance)
        )

    return polygon_gdf




def mask_to_poly_geojson(pred_arr, channel_scaling=None, reference_im=None,
                         output_path=None, output_type='geojson', min_area=40,
                         bg_threshold=0, do_transform=None, simplify=False,
                         tolerance=0.5, **kwargs):
    """Get polygons from an image mask.

    Arguments
    ---------
    pred_arr : :class:`numpy.ndarray`
        A 2D array of integers. Multi-channel masks are not supported, and must
        be simplified before passing to this function. Can also pass an image
        file path here.
    channel_scaling : :class:`list`-like, optional
        If `pred_arr` is a 3D array, this argument defines how each channel
        will be combined to generate a binary output. channel_scaling should
        be a `list`-like of length equal to the number of channels in
        `pred_arr`. The following operation will be performed to convert the
        multi-channel prediction to a 2D output ::

            sum(pred_arr[channel]*channel_scaling[channel])

        If not provided, no scaling will be performend and channels will be
        summed.
    reference_im : str, optional
        The path to a reference geotiff to use for georeferencing the polygons
        in the mask. Required if saving to a GeoJSON (see the ``output_type``
        argument), otherwise only required if ``do_transform=True``.
    output_path : str, optional
        Path to save the output file to. If not provided, no file is saved.
    output_type : ``'csv'`` or ``'geojson'``, optional
        If ``output_path`` is provided, this argument defines what type of file
        will be generated - a CSV (``output_type='csv'``) or a geojson
        (``output_type='geojson'``).
    min_area : int, optional
        The minimum area of a polygon to retain. Filtering is done AFTER
        any coordinate transformation, and therefore will be in destination
        units.
    bg_threshold : int, optional
        The cutoff in ``mask_arr`` that denotes background (non-object).
        Defaults to ``0``.
    simplify : bool, optional
        If ``True``, will use the Douglas-Peucker algorithm to simplify edges,
        saving memory and processing time later. Defaults to ``False``.
    tolerance : float, optional
        The tolerance value to use for simplification with the Douglas-Peucker
        algorithm. Defaults to ``0.5``. Only has an effect if
        ``simplify=True``.

    Returns
    -------
    gdf : :class:`geopandas.GeoDataFrame`
        A GeoDataFrame of polygons.

    """

    mask_arr = preds_to_binary(pred_arr, channel_scaling, bg_threshold)

    if do_transform and reference_im is None:
        raise ValueError(
            'Coordinate transformation requires a reference image.')

    if do_transform:
        with rasterio.open(reference_im) as ref:
            transform = ref.transform
            crs = ref.crs
            ref.close()
    else:
        transform = Affine(1, 0, 0, 0, 1, 0)  # identity transform
        crs = rasterio.crs.CRS()

    mask = mask_arr > bg_threshold
    mask = mask.astype('uint8')

    polygon_generator = features.shapes(mask_arr,
                                        transform=transform,
                                        mask=mask)
    polygons = []
    values = []  # pixel values for the polygon in mask_arr
    for polygon, value in polygon_generator:
        p = shape(polygon).buffer(0.0)
        if p.area >= min_area:
            polygons.append(shape(polygon).buffer(0.0))
            values.append(value)

    polygon_gdf = gpd.GeoDataFrame({'geometry': polygons, 'value': values},
                                   crs=crs.to_wkt())
    if simplify:
        polygon_gdf['geometry'] = polygon_gdf['geometry'].apply(
            lambda x: x.simplify(tolerance=tolerance)
        )
    # save output files
    if output_path is not None:
        if output_type.lower() == 'geojson':
            if len(polygon_gdf) > 0:
                polygon_gdf.to_file(output_path, driver='GeoJSON')
            else:
                save_empty_geojson(output_path, polygon_gdf.crs.to_epsg())
        elif output_type.lower() == 'csv':
            polygon_gdf.to_csv(output_path, index=False)

    return polygon_gdf


def crs_is_metric(gdf):
    """Check if a GeoDataFrame's CRS is in metric units."""
    units = str(gdf_get_projection_unit(gdf)).strip().lower()
    if units in ['"meter"', '"metre"', "'meter'", "'meter'",
                 'meter', 'metre']:
        return True
    else:
        return False


def _check_do_transform(df, reference_im, affine_obj):
    """Check whether or not a transformation should be performed."""
    try:
        crs = getattr(df, 'crs')
    except AttributeError:
        return False  # if it doesn't have a CRS attribute

    if not crs:
        return False  # return False for do_transform if crs is falsey
    elif crs and (reference_im is not None or affine_obj is not None):
        # if the input has a CRS and another obj was provided for xforming
        return True


def instance_mask(df, out_file=None, reference_im=None, geom_col='geometry',
                  do_transform=None, affine_obj=None, shape=(900, 900),
                  out_type='int', burn_value=255, burn_field=None, nodata_value=0):
    """Convert a dataframe of geometries to a pixel mask.

    Arguments
    ---------
    df : :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame`
        A :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame` instance
        with a column containing geometries (identified by `geom_col`). If the
        geometries in `df` are not in pixel coordinates, then `affine` or
        `reference_im` must be passed to provide the transformation to convert.
    out_file : str, optional
        Path to an image file to save the output to. Must be compatible with
        :class:`rasterio.DatasetReader`. If provided, a `reference_im` must be
        provided (for metadata purposes).
    reference_im : :class:`rasterio.DatasetReader` or `str`, optional
        An image to extract necessary coordinate information from: the
        affine transformation matrix, the image extent, etc. If provided,
        `affine_obj` and `shape` are ignored.
    geom_col : str, optional
        The column containing geometries in `df`. Defaults to ``"geometry"``.
    do_transform : bool, optional
        Should the values in `df` be transformed from geospatial coordinates
        to pixel coordinates? Defaults to ``None``, in which case the function
        attempts to infer whether or not a transformation is required based on
        the presence or absence of a CRS in `df`. If ``True``, either
        `reference_im` or `affine_obj` must be provided as a source for the
        the required affine transformation matrix.
    affine_obj : `list` or :class:`affine.Affine`, optional
        Affine transformation to use to convert from geo coordinates to pixel
        space. Only provide this argument if `df` is a
        :class:`geopandas.GeoDataFrame` with coordinates in a georeferenced
        coordinate space. Ignored if `reference_im` is provided.
    shape : tuple, optional
        An ``(x_size, y_size)`` tuple defining the pixel extent of the output
        mask. Ignored if `reference_im` is provided.
    out_type : 'float' or 'int'
    burn_value : `int` or `float`, optional
        The value to use for labeling objects in the mask. Defaults to 255 (the
        max value for ``uint8`` arrays). The mask array will be set to the same
        dtype as `burn_value`. Ignored if `burn_field` is provided.
    burn_field : str, optional
        Name of a column in `df` that provides values for `burn_value` for each
        independent object. If provided, `burn_value` is ignored.
    nodata_value : `int` or `float`, optional
        The value to use for nodata pixels in the mask. Defaults to 0 (the
        min value for ``uint8`` arrays). Used if reference_im nodata value is a float.
        Ignored if reference_im nodata value is an int or if reference_im is not used.
        Take care when visualizing these masks, the nodata value may cause labels to not
        be visualized if nodata values are automatically masked by the software.

    Returns
    -------
    mask : :class:`numpy.array`
        A pixel mask with 0s for non-object pixels and `burn_value` at object
        pixels. `mask` dtype will coincide with `burn_value`.

    """
    # TODO: Refactor to remove some duplicated code here and in other mask fxns

    if out_file and not reference_im:
        raise ValueError(
            'If saving output to file, `reference_im` must be provided.')
    df = _check_df_load(df)

    if len(df) == 0: # for saving an empty mask.
        reference_im = _check_rasterio_im_load(reference_im)
        shape = reference_im.shape
        return np.zeros(shape=shape, dtype='uint8')

    if do_transform is None:
        # determine whether or not transform should be done
        do_transform = _check_do_transform(df, reference_im, affine_obj)

    df[geom_col] = df[geom_col].apply(_check_geom)  # load in geoms if wkt
    if not do_transform:
        affine_obj = Affine(1, 0, 0, 0, 1, 0)  # identity transform

    if reference_im:
        reference_im = _check_rasterio_im_load(reference_im)
        shape = reference_im.shape
        if do_transform:
            affine_obj = reference_im.transform

    # extract geometries and pair them with burn values

    if burn_field:
        if out_type == 'int':
            feature_list = list(zip(df[geom_col],
                                    df[burn_field].astype('uint8')))
        else:
            feature_list = list(zip(df[geom_col],
                                    df[burn_field].astype('float32')))
    else:
        feature_list = list(zip(df[geom_col], [burn_value]*len(df)))

    if out_type == 'int':
        output_arr = np.empty(shape=(shape[0], shape[1],
                                     len(feature_list)), dtype='uint8')
    else:
        output_arr = np.empty(shape=(shape[0], shape[1],
                                     len(feature_list)), dtype='float32')
    # initialize the output array

    for idx, feat in enumerate(feature_list):
        output_arr[:, :, idx] = features.rasterize([feat], out_shape=shape,
                                                   transform=affine_obj)

    if reference_im:
        reference_im = _check_rasterio_im_load(reference_im)
    try:
        bad_data_mask = (reference_im.read() == reference_im.nodata).any(axis=0) # take logical and along all dims so that all pixxels not -9999 across bands
    except AttributeError as ae:  # raise another, more verbose AttributeError
        raise AttributeError("A nodata value is not defined for the source image. Make sure the reference_im has a nodata value defined.") from ae
    if len(bad_data_mask.shape) > 2:
        bad_data_mask = np.dstack([bad_data_mask]*output_arr.shape[2])
        output_arr = np.where(bad_data_mask, 0, output_arr) # mask is broadcasted to filter labels where there are non-nan image values

    if out_file:
        meta = reference_im.meta.copy()
        meta.update(count=output_arr.shape[-1])
        if out_type == 'int':
            meta.update(dtype='uint8')
            if isinstance(meta['nodata'], float):
                meta.update(nodata=nodata_value)
        with rasterio.open(out_file, 'w', **meta) as dst:
            for c in range(1, 1 + output_arr.shape[-1]):
                dst.write(output_arr[:, :, c-1], indexes=c)
            dst.close()

    return output_arr

def geojsons_to_masks_and_fill_nodata(rtiler, vtiler, label_tile_dir, fill_value=0):
    """
    Converts tiled vectors to raster labels and fills nodata values in raster and vector tiles.

    This function must be run after a raster tiler and vector tiler have already been initialized
    and the `.tile()` method for each has been called to generate raster and vector tiles.
    Geojson labels are first converted to rasterized masks, then the labels are set to 0
    where the reference image, the corresponding image tile, has nodata values. Then, nodata
    areas in the image tile are filled  in place with the fill_value. Only works for rasterizing
    all geometries as a single category with a burn value of 1. See test_tiler_fill_nodata in
    tests/test_tile/test_tile.py for an example.

    Args
    -------
    rtiler : RasterTiler
        The RasterTiler that has had it's `.tile()` method called.
    vtiler : VectorTiler
        The VectorTiler that has had it's `.tile()` method called.
    label_tile_dir : str
        The folder path to save rasterized labels. This is created if it doesn't already exist.
    fill_value : str, optional
        The value to use to fill nodata values in images. Defaults to 0.

    Returns
    -------
    rasterized_label_paths : list
        A list of the paths to the rasterized instance masks.
    """
    rasterized_label_paths = []
    print("starting label mask generation")
    if not os.path.exists(label_tile_dir):
        os.mkdir(label_tile_dir)
    for img_tile, geojson_tile in tqdm(zip(sorted(rtiler.tile_paths), sorted(vtiler.tile_paths))):
        fid = os.path.basename(geojson_tile).split(".geojson")[0]
        rasterized_label_path = os.path.join(label_tile_dir, fid + ".tif")
        rasterized_label_paths.append(rasterized_label_path)
        gdf = gpd.read_file(geojson_tile)
        # gdf.crs = rtiler.raster_bounds_crs # add this because gdfs can't be saved with wkt crs
        arr = instance_mask(gdf, out_file=rasterized_label_path, reference_im=img_tile,
                                        geom_col='geometry', do_transform=None,
                                        out_type='int', burn_value=1, burn_field=None) # this saves the file, unless it is empty in which case we deal with it below.
        if not arr.any(): # in case no instances in a tile we save it with "empty" at the front of the basename
            with rasterio.open(img_tile) as reference_im:
                meta = reference_im.meta.copy()
                reference_im.close()
            meta.update(count=1)
            meta.update(dtype='uint8')
            if isinstance(meta['nodata'], float):
                meta.update(nodata=0)
            rasterized_label_path = os.path.join(label_tile_dir, "empty_" + fid + ".tif")
            with rasterio.open(rasterized_label_path, 'w', **meta) as dst:
                dst.write(np.expand_dims(arr, axis=0))
                dst.close()
    rtiler.fill_all_nodata(nodata_fill=fill_value)
    return rasterized_label_paths


############################################################################
# GEOTRANSFORMS FOR GEODF
############################################################################

def georegister_px_df(df, im_path=None, affine_obj=None, crs=None,
                      geom_col='geometry', precision=None, output_path=None):
    """Convert a dataframe of geometries in pixel coordinates to a geo CRS.

    Arguments
    ---------
    df : :class:`pandas.DataFrame`
        A :class:`pandas.DataFrame` with polygons in a column named
        ``"geometry"``.
    im_path : str, optional
        A filename or :class:`rasterio.DatasetReader` object containing an
        image that has the same bounds as the pixel coordinates in `df`. If
        not provided, `affine_obj` and `crs` must both be provided.
    affine_obj : `list` or :class:`affine.Affine`, optional
        An affine transformation to apply to `geom` in the form of an
        ``[a, b, d, e, xoff, yoff]`` list or an :class:`affine.Affine` object.
        Required if not using `raster_src`.
    crs : valid CRS `str`, `int`, or :class:`rasterio.crs.CRS` instance
        The coordinate reference system for the output GeoDataFrame as an EPSG
        code integer. Required if not providing a raster image to extract the
        information from.
    geom_col : str, optional
        The column containing geometry in `df`. If not provided, defaults to
        ``"geometry"``.
    precision : int, optional
        The decimal precision for output geometries. If not provided, the
        vertex locations won't be rounded.
    output_path : str, optional
        Path to save the resulting output to. If not provided, the object
        won't be saved to disk.

    """
    if im_path is not None:
        im = _check_rasterio_im_load(im_path)
        affine_obj = im.transform
        crs = im.crs
    else:
        if not affine_obj or not crs:
            raise ValueError('If an image path is not provided, '
                             'affine_obj and crs must be.')
    crs = _check_crs(crs)
    tmp_df = affine_transform_gdf(df, affine_obj, geom_col=geom_col,
                                  precision=precision)
#    result = gpd.GeoDataFrame(tmp_df, crs='epsg:' + str(crs.to_epsg()))
    result = gpd.GeoDataFrame(tmp_df)
    result = result.set_crs(crs, allow_override=True)

    if output_path is not None:
        if output_path.lower().endswith('json'):
            result.to_file(output_path, driver='GeoJSON')
        else:
            result.to_csv(output_path, index=False)

    return result




#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################




import numpy as np
import skfmm
from osgeo import gdal, ogr

# adding logging
import logging

from .img_utils import get_binary_mask_from_layer
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def GetTruncatedSDT(layer, src_ds, m_per_pixel, clip_vals):
    '''
    Calculate the (unquantized) SDT from a binary mask created from an OGR layer.

    :param ogr.Layer layer: ogr (buildings) layer from geojson or shapefile
    :param GdalDataset src_ds: raster matching input vector layer (SDT will match its shape)
    :param m_per_pixel float: meters per pixel of src_ds
    :param tuple clip_vals: clip SDT values outside this range to the range's edges

    returns: np.ndarray containing the truncated (float) SDT, pre-quantization.
    '''

    # 2022.02.16 refactored to use img_utils.get_binary_mask_from_layer()
    bin_mask = get_binary_mask_from_layer(layer, src_ds)
    
    # Create mask using signed distance function    
    fmm_arr = np.where(bin_mask >= 1, 1, -1)  # Set all values to 1 or -1

    # constant arrays have no zero contour
    # if fmm_arr is constant, return an array with 
    # the minimum possible value of distance 
    # so it has no impact on the final truncated SDT
    if np.all(fmm_arr == fmm_arr[0,0]):
        return np.full(bin_mask.shape, clip_vals[0], dtype=np.int32)
        
    # call skfmm.distance() on it to get a signed distance transform.
    # truncate the SDT to the clip range.
    dist_arr = skfmm.distance(fmm_arr, dx=m_per_pixel).clip(clip_vals[0],clip_vals[1])
        
    # return the (unquantized) signed distance transform array
    return dist_arr

    
    


def QuantizeSDT(the_sdt):
    '''
    quantizes the SDT and returns an Int32 segmentation mask.
    '''
    return np.rint(the_sdt).astype(np.int32)


def CreateModifiedSDTMask(layer, src_ds, m_per_pixel, clip_vals):
    '''
    Creates a Truncated SDT that treats buildings with a common edge as separate buildings.
    :param ogr.Layer layer: ogr (buildings) layer from geojson or shapefile
    :param GdalDataset src_ds: raster matching input vector layer (the returned SDT will match its shape)
    '''
    
    # create tensor to hold results
    num_features = layer.GetFeatureCount()

    log.info(f'Number of features in layer: {num_features}')
    sz_x, sz_y = src_ds.RasterXSize, src_ds.RasterYSize

    if num_features == 0:
        # return a blank mask
        return np.zeros((sz_x, sz_y)).astype(np.int32)

    sdt_array = np.zeros((sz_x, sz_y, num_features)).astype(np.int32)

    # layer is an iterator with mysterious OGR behaviors. 
    # I reset it alot to be sure it is pointed at the head of the feature list.
    layer.ResetReading()

    # make an editable copy of the layer
    # this as usual is a pita in OGR. 
    # Look at this webpage for details: 
    # http://geoexamples.blogspot.com/2012/01/creating-files-in-ogr-and-gdal-with.html
    drv = ogr.GetDriverByName('Memory')
    ram_ds = drv.CreateDataSource( "out" )
    layer_copy = ram_ds.CopyLayer(layer, "copy")
    layer.ResetReading()

    # remove all features in the layer copy
    for feature in layer_copy:
        layer_copy.DeleteFeature(feature.GetFID())

    # check that layer_copy is now empty
    # print(f"Layer_copy feature count: {layer_copy.GetFeatureCount()}")

    # iterate over features in original layer, creating the SDT
    findex = 0

    for feature in layer:
        # create a truncated SDT with the single feature and store it in the ndarray
        layer_copy.CreateFeature(feature)
        sdt_array[:,:,findex] = QuantizeSDT(GetTruncatedSDT(layer_copy, src_ds, m_per_pixel, clip_vals))
        # print(f"Debug: {np.max(sdt_array[:,:,findex])}")
        findex = findex + 1
        layer_copy.DeleteFeature(feature.GetFID())
    layer.ResetReading()

    # take the pixelwise max over the stack of SDTs and return it
    final_SDT = sdt_array.max(axis=2)

    # clean up 
    ram_ds = None

    return final_SDT

def ProcessLabelToSdtMask(json_path, src_ds, mpp, clip_vals):
    
    log.info(f"Processing label file: {json_path}")
    drv = ogr.GetDriverByName('GeoJSON')
    ogr_ds = drv.Open(json_path)
    lyr = ogr_ds.GetLayer()

    # create the truncated signed distance transform mask
    sdt_array = CreateModifiedSDTMask(lyr, src_ds, mpp, clip_vals)

    # added 20220520: adjust mask to have values in the range [0,numclasses)
    sdt_array = sdt_array - clip_vals[0]
    return sdt_array



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import glob, os
from unittest import result

def locate_field(e, field_name):
    item = e.get(field_name, None)
    if item is not None:
        return item
    else:
        for key in e.keys():
            item = e[key]
            if isinstance(item, dict):
                # print(f"{key} points to a dictionary")
                result = locate_field(item, field_name)
                if result is not None:
                    return result
    return None

# Names of standard hyperparameters
# for use in experiment logs

EXP_NOTES = 'EXP_NOTES'
N_EPOCHS = 'N_EPOCHS'
BATCH_SIZE = 'BATCH_SIZE'
LEARNING_RATE = 'LEARNING_RATE'
LOSS_FN = 'LOSS_FN'
OPTIMIZER = 'OPTIMIZER'
TB_LOGDIR = 'TB_LOGDIR'
TIMESTAMP = 'TIMESTAMP'
TRAINING_IMG_DIR = 'TRAINING_IMG_DIR'
TRAINING_MASK_DIR = 'TRAINING_MASK_DIR'
VAL_IMG_DIR = 'VAL_IMG_DIR'
VAL_MASK_DIR = 'VAL_MASK_DIR'
OUTPUT_IMG_SIZE = 'OUTPUT_IMG_SIZE'
MODEL_FN = 'MODEL_FN'
RANDOM_SEED = 'RANDOM_SEED'
N_CLASSES = 'N_CLASSES'
BEST_MODEL_VALUE = 'BEST_MODEL_VALUE'
BEST_MODEL_EPOCH = 'BEST_MODEL_EPOCH'
KEEPER = 'KEEPER'
ENCODER_FREEZE = 'ENCODER_FREEZE'


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import os, sys, re, shutil
from random import sample as rsample
from pathlib import Path
import pandas as pd

def get_basename(filepath):
    '''
    recovers the  basename of an image, mask, prediction, or label file 
    in any location in a RAMP training or output dataset.
    i.e., basename.mask.tif -> basename
    basename.tif -> basename
    basename.mask.sdt.tif -> basename
    basename.pred.tif -> basename
    basename.geojson -> basename
    '''
    # the basename ends at the string '.mask' or '.pred'
    # the stem will be basename or basename.mask or basename.mask.sdt 
    filepath = str(filepath)
    the_stem = Path(filepath).stem
    rex = r"(?P<basename>.*)((\.mask)|(\.pred))"
    reg = re.compile(rex)
    m = reg.search(the_stem)
    if m is None:
        # the 'mask' and 'predict' parts didn't match
        return the_stem
    else:
        return m['basename']
 

def get_matching_pairs(d1_iter, d2_iter):
    '''
    Given 2 lists of files, 
    constructs an iterator over matching pairs left-joined on the first list.
    Pairs are matched if get_basename(file1) == get_basename(file2). 
    '''    
    d1_files = [f for f in d1_iter]
    d2_files = [f for f in d2_iter]

    num_d1_files = len(d1_files)

    # sanity check
    assert len(d1_files) == len(d2_files),\
        "Directories contain unequal counts of chip and label files"

    d1_basenames = [get_basename(fn) for fn in d1_files]
    d2_basenames = [get_basename(fn) for fn in d2_files]

    df1 = pd.DataFrame(
        {
            "filename_1":d1_files,
            "basename":[get_basename(fn) for fn in d1_files]
        }
    )
    df2 = pd.DataFrame(
        {
            "filename_2":d2_files,
            "basename":[get_basename(fn) for fn in d2_files]
        }
    )

    joined_df = pd.merge(df1, df2, on="basename", how="left")
    paired_files = [(row["filename_1"], row["filename_2"]) for index, row in joined_df.iterrows()]
    assert len(paired_files) == num_d1_files, \
        "Directories contain mismatched chip and label files"
    return paired_files

def get_SRS_files(sample_size, the_dir, pattern='**/*'):
    '''
    returns a simple random sample (no repeated files) from the directory.
    if sample_size > number files in directory, returns complete list of files. 
    pattern: string to pass to the_dir.glob(pattern) to get the list of files. Default is trivial pattern.  
    '''

    dirpath = Path(the_dir)
    assert dirpath.is_dir(), f"Directory {the_dir} doesn't exist"
    files = dirpath.glob(pattern)

    # select only files, no directories
    filenames = [fn for fn in files if fn.is_file()]

    # return full list
    numfiles = len(filenames)
    if sample_size >= numfiles:
        return files

    file_sample = rsample(filenames, sample_size)
    return file_sample


def move_training_data_files(basename_list, src_dir, target_dir, move_file=False):
    '''
    # todo: change print statements to log statements
    basename_list: list of files to take basenames from, usually from get_SRS_files
    move_file: if true, move the file: if false, copy it
    '''    
    # Create target dirs if they don't exist
    target_path = Path(target_dir)
    target_path.mkdir(parents=True, exist_ok=True)
            
    # get list of basenames from basename_list for files to move
    basenames = []
    for filename in basename_list:
            # get the 'true stem' for files with >1 suffix: 
            # i.e., it handles names like basename.mask.tif correctly
            basename = get_basename(filename)
            basenames.append(basename)

    # copy over files from the source having basenames in 'basenames' list
    src_files_list = Path(src_dir).glob('**/*')
    for sfn in src_files_list:
        sbase = get_basename(sfn)
        if sbase in basenames:
            if move_file:
                #print(f"Move: {sfn} to {target_dir}")
                shutil.move(str(sfn), target_dir)
            else:
                #print(f"Copy: {sfn} to {target_dir}")
                shutil.copy(str(sfn), target_dir)



def get_matching_mask(basename, mask_dir):
    '''
    return full path to 'basename.mask.tif' in mask_dir
    '''
    mdir = Path(mask_dir)
    return mdir / f"{basename}.mask.tif"
    
    


    




#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import tensorflow as tf
# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def get_sparse_categorical_accuracy_fn(cfg):
    return tf.keras.metrics.SparseCategoricalAccuracy()


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

import os, re

def construct_mask_filepath(mask_dir, path_to_chip):
    '''
    Constructs a standard mask file path given the mask directory and the chip filename.
    Standard masks have the name: (chip_basename).mask.tif.
    This means they will alphabetically sort the same way as the image chips did.
    '''

    chip_filename = os.path.split(path_to_chip)[1]
    chip_basename = os.path.splitext(chip_filename)[0]
    mask_file = chip_basename + '.mask.tif'
    return os.path.join(mask_dir, mask_file)

def get_tq_chip_label_pairs(chip_dir, label_dir):
    '''
    Get a list of pairs of matching chips and labels from directories.
    Checks to be sure all chips matching delivered labels are present.

    Matched files are named (e.g.) 84930efd.tif and 84930efd.geojson.

    :param directory chip_dir: directory containing image chips
    :param directory label_dir: directory containing geojson vectors

    Returns: list of tuples of chip and label filename pairs.
    '''

    # construct list of matching image chips from json labels

    # get sorted geojson file names
    gj_files = sorted([gj_file for gj_file in os.listdir(label_dir) if gj_file.endswith('.geojson')])
    
    # get matching chip file names -- for TQ deliveries these have the same base filename 
    basenames = [os.path.splitext(gj_file)[0] for gj_file in gj_files]
    # print(basenames[0])
    chip_files = [basename + '.tif' for basename in basenames]

    # get full pathnames to json and chip files
    json_paths = [os.path.join(label_dir, gj_file) for gj_file in gj_files]
    chip_paths = [os.path.join(chip_dir, chip_file) for chip_file in chip_files]

    # check to be sure the matching image chips are present
    for chip_path in chip_paths:
        if not os.path.exists(chip_path):
            raise FileNotFoundError(f'Missing: required image chip file {chip_path}')
    
    return list(zip(chip_paths, json_paths))




import os

from setuptools import find_packages, setup


def readme() -> str:
    """Utility function to read the README.md.

    Used for the `long_description`. It's nice, because now
    1) we have a top level README file and
    2) it's easier to type in the README file than to put a raw string in below.

    Args:
        nothing

    Returns:
        String of README.md file.
    """
    return open(os.path.join(os.path.dirname(__file__), "README.md")).read()


setup(
    name="ramp-fair",
    version="0.1.1",
    author="Carolyn Johnston",
    author_email="carolyn.johnston@dev.global",
    description="Replicable AI for Microplanning , Fork maintained for fAIr",
    python_requires=">=3",
    license="",
    url="",
    packages=find_packages(),
    package_dir={"": "."},
    long_description=readme(),
    long_description_content_type="text/markdown",
)



#################################################################
#
# created for ramp project, August 2022
#
#################################################################



# base class for all custom exceptions in Ramp code
class RampError(Exception):
    pass


# thrown when gdal.Open fails
class GdalReadError(RampError):
    pass



from . import chip_utils
from . import eval_utils
from . import file_utils
from . import geo_utils
from . import img_utils
from . import imgproc_utils
from . import label_utils
from . import log_fields
from . import lrfinder
from . import mask_to_vec_utils
from . import misc_ramp_utils
from . import model_utils
from . import multimask_utils
from . import ramp_exceptions
from . import sdt_mask_utils


#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################



import albumentations as A
import cv2
import re

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def get_augmentation_fn(cfg):
    aug_name_list = cfg["augmentation"]["aug_list"]
    aug_parms_list = cfg["augmentation"]["aug_parms"]

    # construct augmentation functions
    aug_fn_list = []
    for augment_name, aug_parms in zip(aug_name_list, aug_parms_list):
        aug_fn = getattr(A, augment_name)

        # some of these values are cv2 flag strings that have to be converted
        flag_parms = { k: getattr(cv2,v) for (k,v) in aug_parms.items() 
            if (isinstance(v, str) and hasattr(cv2, v))}
        aug_parms = {k: v for (k,v) in aug_parms.items() if not isinstance(v, str)}
        aug_parms.update(flag_parms)
        aug_fn_list.append(aug_fn(**aug_parms))
    return A.Compose(aug_fn_list)

        



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################

# adding logging
import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

import tensorflow as tf
import keras.backend as K
from tensorflow.keras.losses import CategoricalCrossentropy, categorical_crossentropy

import numpy as np


def get_sparse_categorical_crossentropy_fn(cfg):
    '''
    returns sparse categorical cross entropy. Does not use parameters. 
    '''
    return tf.keras.losses.sparse_categorical_crossentropy

def get_custom_loss_fn(cfg):    
    '''
    returns the Google Africa custom loss function with given parameters.
    sigma: length scale for the gaussian-filtered weight image in the weighted scce loss. 
    myconst: constant multiple for the gaussian-filtered weight image in the weighted scce loss.
    gamma (real, >0): FTL power parameter.
    alpha (real, 0-1): trades off cost of FP vs FN. alpha larger => FP counts more.
    epsilon (real,>0): small positive number to stabilize FTL computations.
    ftl_wt (real): trades off FT loss vs wtd scce loss. flt_wt larger => FT loss counts more.
    '''           
    
    ftl_wt = cfg["loss"]["loss_fn_parms"]["ftl_wt"]

    def custom_loss_fn(y_true, y_hat):
        wscce_loss = get_weighted_SCCE_loss_fn(cfg)
        ft_loss = get_focal_tversky_loss_function(cfg)
        # return wscce_loss(y_true, y_hat) + 
        return ftl_wt*ft_loss(y_true, y_hat)
    return custom_loss_fn



def get_weighted_SCCE_loss_fn(cfg):
    '''
    Returns a sparse categorical cross entropy loss function with per-pixel weighting.
    '''
    from ..utils.imgproc_utils import batch_get_weight_image
    num_classes = cfg["num_classes"]
    sigma = cfg["loss"]["loss_fn_parms"]["sigma"]
    myconst = cfg["loss"]["loss_fn_parms"]["myconst"]
    assert num_classes == 2, "Multiclass (>2) masks are not currently supported"

    def weighted_SCCE_loss(y_true, y_hat):

        xentropy_loss = CategoricalCrossentropy()

        # get the weighting image from the mask. 
        # it has shape: (B,H,W)
        wt_img = batch_get_weight_image(y_true, sigma, myconst)

        # one hot encode and squeeze y_true to have shape (B, H, W, num_classes)
        yt = tf.squeeze(tf.one_hot(tf.cast(y_true, tf.int32), num_classes))
        return xentropy_loss(yt, y_hat, sample_weight=wt_img)

    return weighted_SCCE_loss



def get_SCCE_labelsmoothing_loss_fn(cfg):
    #num_classes, ls_param=0.1):
    '''
    Returns a sparse categorical cross entropy loss function that applies label smoothing.
    '''
    num_classes = cfg["num_classes"]
    ls_param = cfg["loss"]["loss_fn_parms"]["ls_param"]
    def SCCE_labelsmoothing(y_true, y_hat):
        yt = tf.squeeze(tf.one_hot(tf.cast(y_true, tf.int32), num_classes))
        return categorical_crossentropy(yt, y_hat, label_smoothing = ls_param)

    return SCCE_labelsmoothing


def get_class_tversky_fn(alpha, epsilon):
    
    def class_tversky(y_true, y_pred):
        '''
        Returns a vector with length equal to the number of classes.
        Each class contributes to the total FTL in a 1-vs-rest manner.
        '''

        # Note: batch_flatten actually flattens all the other dimensions. 
        # So, for example, a (3, 5, 6, 2)-tensor becomes a (3, 60)-tensor.
        # In this case, because you permuted channels and batches, you 
        # get a (C, HWB)-tensor.
        y_true_pos = K.batch_flatten(y_true)
        y_pred_pos = K.batch_flatten(y_pred)

        # Note that each of these sums is over the HWB dimension, so you 
        # get a C-vector in each case. 
        # TP(c): terms are 0 if true class isn't c, large if the true class was 
        # c and you got it right.
        true_pos = K.sum(y_true_pos * y_pred_pos, 1)
        
        # FN(c): terms are 0 if true class isn't c, large if the true class was 
        # c and you got it wrong.
        false_neg = K.sum(y_true_pos * (1-y_pred_pos), 1)
        
        # FP(c): terms are 0 if true class is c. and large if the true class 
        # wasn't c, but you guessed that it was.
        false_pos = K.sum((1-y_true_pos)*y_pred_pos, 1)
        
        # alpha trades off FP and FN.
        # epsilon stabilizes the division.
        return (true_pos + epsilon)/(true_pos + alpha*false_neg + (1-alpha)*false_pos + epsilon)

    return class_tversky

def get_focal_tversky_loss_function(cfg):
    #num_classes, gamma, alpha, epsilon):
    '''
    returns a FTL function with the given parameters.
    num_classes (integer): # classes in the mask. 
    gamma (real): FTL main parameter.
    alpha (real): trades off cost of FP vs FN.
    epsilon (real): small factor to stabilize FTL computations.
    '''
    
    num_classes = cfg["num_classes"]
    gamma = cfg["loss"]["loss_fn_parms"]["gamma"]
    alpha = cfg["loss"]["loss_fn_parms"]["alpha"]
    epsilon = cfg["loss"]["loss_fn_parms"]["epsilon"]

    class_tversky = get_class_tversky_function(alpha, epsilon)

    def focal_tversky_loss(y_true, y_pred):
        '''
        returns the sum of the classwise losses, raised to the gamma
        power to make it nonlinear.
        '''
        # y_true and y_pred must be one-hot encoded. 
        # usually, y_true is sparse-encoded, and y_pred is one-hot encoded. 
        # so transform y_true first.
        yt = tf.one_hot(tf.cast(tf.squeeze(y_true, 3), tf.int32), num_classes)

        # for tensors of shape batch-H-W-channels, this trades
        # the batch and channels dims, so you get a channels-H-W-batch shape.
        yt = K.permute_dimensions(yt, (3,1,2,0))
        y_pred = K.permute_dimensions(y_pred, (3,1,2,0))
        
        # calculate classwise components of FTL
        pt_1 = class_tversky(yt, y_pred)
        return K.sum(K.pow((1-pt_1), gamma))

    return focal_tversky_loss


def main():

    # unit test
    num_classes = 3
    y_true = np.array((0,1,2,0,1,2,0,1,2))

    # this guesses 0 a lot, and 2, hardly ever. 
    y_pred_0 = np.array((0.25, 0.25, 0.5, 0.5, 0.5, 0.5, 0.5, 0.25, 0.25))
    y_pred_1 = np.array((0.5, 0.5, 0.25, 0.25, 0.25, 0.25, 0.25, 0.5, 0.25))
    y_pred_2 = np.array((0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.5))

    # shape is (9, 3) -- classes last as comes out of the NN
    y_pred = np.array((y_pred_0, y_pred_1, y_pred_2)).astype('float32').transpose()
    

    # one hot encode the truth vector
    yt = tf.squeeze(tf.one_hot(tf.cast(y_true, tf.int32), num_classes))
    
    yt = K.permute_dimensions(yt, (1,0))
    yp = K.permute_dimensions(y_pred, (1,0))

    class_tversky = get_class_tversky_function(alpha=0.5, epsilon=0)
    CT = class_tversky(yt, yp)
    truth = np.array((0.38462, 0.41667,0.36364))
    assert np.sum(np.square(CT - truth)) < 10E-06, f"returned unexpected value from class_tversky(): {CT}"

    FT_loss = get_focal_tversky_loss_function(3, 1.0, 0.5, 0.0)
    # reshape y_true and y_pred so they look like tf tensors

    y_true = y_true.reshape((1, 3, 3))
    y_pred = y_pred.reshape((1, 3, 3, 3))

    FTL = FT_loss(y_true, y_pred)
    truth_ftl = 1.83507
    print(FTL)
    assert np.square(FTL - truth_ftl) < 10E-06, f"returned unexpected FTL: {FTL}"
    

    return


    





if __name__=='__main__':
    main()



#################################################################
#
# created for ramp project, August 2022
# Author: carolyn.johnston@dev.global
#
#################################################################


import sys, os, re, csv
from pathlib import Path

# Filter functions for csv

def filter_accepted_labels(csvfile, fn_pos=0, accepted_pos=-1, filter_rejected=False):
    '''
    create a list of accepted label filenames from a chip_review.json file created by chippy checker
    existing_csv: list of csv rows returned by csv.Reader
    fn_pos: position of the filename column in csvfile (default 0)
    filter_rejected: set to 'True' if you want to create a list of rejected labels
    '''
    filtered_labels = []
    with open(csvfile, 'r') as fileh:
        reader = csv.reader(csvfile, delimiter = ',')
        for row in reader:
            filename = Path(row[fn_pos])
            accepted = bool(row[accepted_pos])

        condition = accepted if (not filter_rejected) else (not accepted)
        if condition:
            filtered_labels.append(str(filename))
    return filtered_labels

def get_lonlat_from_filename(i_filename):
    '''
    get lon lat (or easting northing) of the upper left hand corner (I think) from 
    chip filenames created by the solaris tiler
    example: 'sn_paris_121_2.212736_49.020317.geojson' returns (2.212736, 49.020317)
    
    i_filename: the filename
    returns: (lon, lat) or None
    '''
    try:
        basename = Path(i_filename).stem
        re_str = r".*_(?P<lon>\-?[0-9]+\.?[0-9]+)_(?P<lat>\-?[0-9]+\.?[0-9]+)$" 
        rex = re.compile(re_str)
        match = rex.search(basename)
        lat = float(match.group('lat'))
        lon = float(match.group('lon'))
        return (lon, lat)
    except:
        return None

